# SPDX-License-Identifier: MIT
"""URL - Vlju representable as a URL."""

from fnattr.vlju.types.uri import URI

class URL(URI):
    """Represents a URL."""

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URL:
            return (self._value,
                    {
                        'scheme': self._scheme,
                        'authority': self._authority,
                        'query': self._query,
                        'fragment': self._fragment,
                        'urnq': self._urnq,
                        'urnr': self._urnr,
                        'sa': self._sa,
                        'ap': self._ap,
                    })
        return super().cast_params(t)

    def url(self) -> str:
        return self.lv()



# SPDX-License-Identifier: MIT
"""Kinds of info: URI."""

# https://tools.ietf.org/html/rfc4452

import fnattr.vlju.types.doi
import fnattr.vlju.types.lccn

# yapf: disable
KIND = {
    'ark':                  None,
    'arxiv':                None,
    'bibcode':              None,
    'bnf':                  None,
    'ddbj-embl-genbank':    None,
    'dlf':                  None,
    'doi':                  fnattr.vlju.types.doi.DOI,
    'eu-repo':              None,
    'fedora':               None,
    'hdl':                  fnattr.vlju.types.doi.DOI,
    'inchi':                None,
    'lanl-repo':            None,
    'lc':                   None,
    'lccn':                 fnattr.vlju.types.lccn.LCCN,
    'ncid':                 None,
    'netref':               None,
    'nla':                  None,
    'nyu':                  None,
    'oclcnum':              None,
    'ofi':                  None,
    'pmid':                 None,
    'pronom':               None,
    'refseq':               None,
    'rfa':                  None,
    'rlgid':                None,
    'sici':                 None,
    'sid':                  None,
    'srw':                  None,
    'ugent-repo':           None,
}
# yapf: enable

# yapf: disable
ORGANIZATION = {
    'ark':          'Archival Resource Keys',
    'arxiv':        'arXiv.org identifiers',
    'bibcode':      'Astrophysics Data System bibcodes',
    'bnf':          'Bibliothèque nationale de France',
    'ddbj-embl-genbank':    'DDBJ/EMBL/GenBank',
    'dlf':          'Digital Library Federation',
    'doi':          'Digital Object Identifiers',
    'eu-repo':      'European Repository Systems',
    'fedora':       'Fedora Digital Objects and Disseminations',
    'hdl':          'Handles',
    'inchi':        'IUPAC International Chemical Identifiers (InChI)',
    'lanl-repo':    'LANL Research Library',
    'lc':           'Library of Congress Identifiers',
    'lccn':         'Library of Congress Control Numbers',
    'ncid':         'NACSIS-CAT Identification Numbers',
    'netref':       'NISO Standard for Network Reference Services',
    'nla':          'National Library of Australia',
    'nyu':          'New York University Digital Objects',
    'oclcnum':      'OCLC Worldcat Control Numbers',
    'ofi':          'Registry Identifiers in the NISO OpenURL Framework',
    'pmid':         'PubMed',
    'pronom':       'PRONOM Unique Identifiers',
    'refseq':       'identifiers for RefSeq reference sequence record',
    'rfa':          'Registry Framework Architecture',
    'rlgid':        'RLG Database Record',
    'sici':         'Serial Item and Contribution',
    'sid':          'Source Identifiers in the NISO OpenURL Framework',
    'srw':          'Search/Retrieve Web Services',
    'ugent-repo':   'University Library Ghent',
}
# yapf: enable



# SPDX-License-Identifier: MIT
"""Generic web site items."""

import re

from collections.abc import Iterable
from typing import Self

from fnattr.util import fearmat
from fnattr.util.error import Error
from fnattr.vlju.types.uri import URI, Authority
from fnattr.vlju.types.url import URL

Template = str | None

class SiteBase(URL):
    """Base class for SiteFactory-generated Vljus."""

    _scheme: str
    _authority: Authority | None = None
    path_template: Template = None
    query_template: Template = None
    fragment_template: Template = None
    normalize_template: Template = None
    url_patterns: list[tuple[re.Pattern, str]] = []

    def __init__(self, s: str) -> None:
        if (t := match_url(type(self).url_patterns, s)):
            s = t
        if self.normalize_template:
            s = fearmat.fearmat(self.normalize_template, {'id': s, 'x': s})
        super().__init__(s, scheme=self._scheme, authority=self._authority)

    def __str__(self) -> str:
        return self._value

    def path(self) -> str:
        if self.path_template:
            return fearmat.fearmat(self.path_template, {
                'id': self._value,
                'x': self._value,
            })
        return self._value

    def query(self) -> str:
        if self.query_template:
            return fearmat.fearmat(self.query_template, {
                'id': self._value,
                'x': self._value,
            })
        return ''

    def fragment(self) -> str:
        if self.fragment_template:
            return fearmat.fearmat(self.fragment_template, {
                'id': self._value,
                'x': self._value,
            })
        return ''

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URI or t is URL:
            return (self.path(),
                    {
                        'scheme': self._scheme,
                        'authority': self._authority,
                        'query': self.query(),
                        'fragment': self.fragment(),
                    })
        raise self.cast_param_error(t)

    @classmethod
    def from_url(cls, url: str) -> Self:
        if (s := cls.match_url(url)):
            return cls(s)
        message = f'URL not recognized: {url!r}'
        raise Error(message)

    @classmethod
    def match_url(cls, url: str) -> str:
        return match_url(cls.url_patterns, url)

def site_class(name: str,
               host: Authority | str,
               path: Template,
               scheme: str | None = None,
               query: Template = None,
               fragment: Template = None,
               normalize: Template = None,
               url: Iterable[str | list[str]] | None = None) -> type[SiteBase]:
    return type(
        name, (SiteBase, ),
        {
            '_scheme': scheme if scheme is not None else 'https',
            '_authority': Authority(host),
            'path_template': unlistify(path),
            'query_template': unlistify(query),
            'fragment_template': unlistify(fragment),
            'normalize_template': unlistify(normalize),
            'url_patterns': site_url_patterns(url),
        })

def unlistify(s: str | list[str] | None) -> str | None:
    if s is None or isinstance(s, str):
        return s
    return ''.join(s)

def site_url_patterns(
    urls: Iterable[str | list[str]] | None = None,
) -> list[tuple[re.Pattern, str]]:
    if not urls:
        return []
    r = []
    for url in urls:
        if isinstance(url, list):
            pattern = url[0]
            replacement = url[1] if len(url) > 1 else r'\1'
        else:
            pattern = url
            replacement = r'\1'
        r.append((re.compile(pattern, re.ASCII | re.VERBOSE), replacement))
    return r

def match_url(patterns: list[tuple[re.Pattern, str]], url: str) -> str:
    for pattern, replacement in patterns:
        if (m := pattern.fullmatch(url)):
            return m.expand(replacement)
    return ''



# SPDX-License-Identifier: MIT
"""ISSN - International Standard Serial Number."""

from fnattr.util import checksum
from fnattr.vlju.types.ean import EAN13, as13

class ISSN(EAN13):
    """Represents an ISSN (International Standard Serial Number)."""

    def __init__(self, s: str) -> None:
        # self._value contains an unsplit ISSN-13 string.
        v = as13(s, 'issn')
        if v is None:
            raise ValueError(s)
        super().__init__(v, 'issn')

    def lv(self) -> str:
        if (v := self.split8()) is None:    # pragma: no branch
            raise ValueError(v)             # pragma: no cover
        return v

    def path(self) -> str:
        return self.lv()

    def issn13(self) -> str:
        """Return an unsplit ISSN-13."""
        return self._value

    def issn8(self) -> str | None:
        """Return an unsplit ISSN-8, or None if not representable."""
        if self._value and self._value.startswith('977'):
            s = self._value[3 : 10]
            return s + checksum.mod11(s)
        return None

    def split13(self) -> str:
        v = self._value
        return f'{v[0:3]}-{v[3:7]}-{v[7:10]}-{v[10:12]}-{v[12]}'

    def split8(self) -> str | None:
        if (s := self.issn8()) is None:
            return None
        return f'{s[0:4]}-{s[4:8]}'



# SPDX-License-Identifier: MIT
"""Vlju factories."""

from collections.abc import Callable, Mapping
from typing import Self

from fnattr.util.error import Error
from fnattr.vlju import Vlju

VljuFactory = Callable[[str, str], tuple[str, Vlju]]

class FactoryError(Error):
    """Factory error."""

def default_factory(k: str, v: str) -> tuple[str, Vlju]:
    return (k, Vlju(v))

class MappedFactory:
    """VljuFactory that maps keys to Vlju types."""

    def __init__(self,
                 kmap: Mapping[str, type[Vlju]],
                 default: type[Vlju] = Vlju) -> None:
        self.kmap = dict(kmap)
        self.default = default

    def setitem(self, k: str, v: type[Vlju]) -> Self:
        self.kmap[k] = v
        return self

    def __call__(self, k: str, v: str) -> tuple[str, Vlju]:
        try:
            return (k, self.kmap.get(k, self.default)(v))
        except Exception as e:  # noqa: blind-except
            msg = f'{k} {v}'
            raise FactoryError(msg) from e

class LooseMappedFactory(MappedFactory):
    """VljuFactory that maps keys for Vlju types, and reverts to default."""

    def __call__(self, k: str, v: str) -> tuple[str, Vlju]:
        try:
            value = self.kmap.get(k, self.default)(v)
        except Exception:   # noqa: blind-except
            value = self.default(v)
        return (k, value)



# SPDX-License-Identifier: MIT
"""Generate isbn_ranges.py."""

# ruff: noqa: S101

import argparse
import array
import datetime
import logging
import pathlib
import pprint
import sys
import xml.etree.ElementTree as ElT

from collections.abc import MutableSequence
from pathlib import Path
from typing import TextIO

from fnattr.vlju.types.ean.isbn import RangeAgencies, Ranges

class RangesFromXml(Ranges):
    """Load ranges from an XML file."""

    def __init__(self, xmlfile: Path) -> None:
        """Load ISBN ranges from a file."""
        root = ElT.parse(xmlfile).getroot()
        assert root.tag == 'ISBNRangeMessage'

        starts: list[int] = []
        lengths: list[int] = []
        self.agencies: RangeAgencies = {}
        for rg in root.iter('RegistrationGroups'):
            for g in rg.iter('Group'):
                self._load_group(g, starts, lengths, self.agencies)
        super().__init__(array.array('Q', starts), array.array('I', lengths))

    def _load_group(self,
                    g: ElT.Element,
                    starts: MutableSequence[int],
                    lengths: MutableSequence[int],
                    agencies: RangeAgencies) -> None:
        s_prefix = g.findtext('Prefix')
        assert s_prefix is not None
        s_compact_prefix = s_prefix.replace('-', '')
        ts_prefix: tuple[str, ...] = tuple(s_prefix.split('-'))
        agency = g.findtext('Agency')
        if agency:
            agencies[ts_prefix] = agency
        ti_prefix_length: tuple[int, ...] = tuple(map(len, ts_prefix))
        remaining = 12 - sum(ti_prefix_length)
        logging.debug('prefix %s %s + %d',
                      s_prefix,
                      ti_prefix_length,
                      remaining)

        for r in g.iter('Rule'):
            s_range = r.findtext('Range')
            assert s_range is not None
            s_length = r.findtext('Length')
            assert s_length is not None
            length = int(s_length)

            s_first, s_last = s_range.split('-')
            pad = '0' * (remaining - len(s_first))
            s_first = (s_first + pad)[: remaining]
            s_last = (s_last + pad)[: remaining]

            if length:
                ti_lengths = (*ti_prefix_length, length, remaining - length, 1)
            else:
                ti_lengths = (*ti_prefix_length, remaining, 1)
            assert sum(ti_lengths) == 13
            s_lengths = ''.join(map(str, ti_lengths))
            # Assert that each segment length is at most 9.
            assert len(s_lengths) == len(ti_lengths)
            i_lengths = int(s_lengths[1 :])

            s_compact_first = s_compact_prefix + s_first + '0'
            s_compact_last = s_compact_prefix + s_last + '9'

            i_first = int(s_compact_first)
            i_last = int(s_compact_last)

            logging.debug('B %d to %d : %s to %s : %s %d',
                          i_first,
                          i_last,
                          '-'.join(self._isplit(s_compact_first, i_lengths)),
                          '-'.join(self._isplit(s_compact_last, i_lengths)),
                          ti_lengths,
                          i_lengths)

            starts.append(i_first)
            lengths.append(i_lengths)

    def write(self, f: TextIO) -> None:
        f.write('\nAGENCY = ')
        pprint.pprint(self.agencies, stream=f)
        f.write(f'\nSTART = {self._start!r}\n')
        f.write(f'\nSPLIT = {self._split!r}\n')

def main(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv
    cmd = pathlib.Path(argv[0]).stem
    parser = argparse.ArgumentParser(
        prog=cmd, description='Generate ISBN ranges')
    parser.add_argument(
        '--input',
        '-i',
        metavar='FILE',
        default='third_party/data/RangeMessage.xml',
        help='Input file.')
    parser.add_argument(
        '--output',
        '-o',
        metavar='FILE',
        default='src/vlju/types/ean/isbn_ranges.py',
        help='Output file.')
    args = parser.parse_args(argv[1 :])

    ranges = RangesFromXml(args.input)
    with pathlib.Path(args.output).open('w', encoding='utf-8') as f:
        f.write('"""Generated ISBN range data."""\n\n')
        f.write('# DO NOT EDIT!\n')
        f.write(f'# Generated from {args.input}\n')
        f.write(
            f'# at {datetime.datetime.now(tz=datetime.UTC).isoformat()}\n\n')
        f.write('from array import array\n')
        ranges.write(f)

    return 0

if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: MIT
"""ISBN (International Standard Book Number)."""

import array
import bisect
import warnings

from fnattr.util import checksum
from fnattr.vlju.types.ean import EAN13, as13, is_valid_ean13, isbn_ranges

def constraint(t: object, s: str) -> None:
    if not t:                   # pragma: no branch
        raise RuntimeError(s)   # pragma: no cover

RangeAgencies = dict[tuple[str, str], str]

class Ranges:
    """Provides ISBN splitting."""

    def __init__(self, start: array.array, split: array.array) -> None:
        # SoA: _start is sorted lower bounds, _split is corresponding split.
        self._start = start
        self._split = split

    def split(self, s: str) -> tuple[str, ...]:
        """Split a string isbn1, with no separators."""
        constraint(len(s) == 13, 'length not 13')
        constraint(s.isdigit(), 'non-digit')
        i = bisect.bisect_right(self._start, int(s))
        if i <= 0:
            warnings.warn(f'ISBN {s} not found in split table', stacklevel=0)
            return (s, )
        return self._isplit(s, self._split[i - 1])

    @staticmethod
    def _isplit(s: str, i: int) -> tuple[str, ...]:
        """Split a string according to the integer pattern."""
        r = []
        while i:
            d = int(i % 10)
            i = i // 10
            constraint(d != 0, 'zero segment')
            r.append(s[-d :])
            s = s[:-d]
        r.append(s)
        return tuple(reversed(r))

class ISBN(EAN13):
    """Represents an ISBN (International Standard Book Number)."""

    _ranges = Ranges(isbn_ranges.START, isbn_ranges.SPLIT)
    split_all = False

    def __init__(self, s: str, *, split: bool = False) -> None:
        # self._value contains an unsplit ISBN-13 string.
        v = as13(s, 'isbn')
        if v is None:
            raise ValueError(s)
        super().__init__(v, 'isbn')
        self._parts: tuple[str, ...] | None = None
        if split or self.split_all:
            self.split()

    def isbn13(self) -> str:
        """Return an unsplit ISBN-13."""
        return self._value

    def isbn10(self) -> str | None:
        """Return an unsplit ISBN-10, or None if not representable."""
        if self._value and self._value.startswith('978'):
            s = self._value[3 : 12]
            return s + checksum.mod11(s)
        return None

    def split(self) -> tuple[str, ...]:
        if self._parts is None:
            self._parts = self._ranges.split(self._value)
        return self._parts

    def split13(self) -> str:
        """Return a canonically split ISBN-13."""
        return '-'.join(map(str, self.split()))

    def split10(self) -> str | None:
        """Return a canonically split ISBN-10."""
        parts = self.split()
        if parts[0] == '978':
            check = checksum.mod11(self._value[3 : 12])
            return '-'.join(parts[1 :-1]) + '-' + check
        return None

def is_valid_isbn10(s: str) -> bool:
    """Check for 10-character-only form."""
    return (len(s) == 10 and s[0 : 8].isdigit()
            and checksum.mod11(s[0 : 8]) == s[9])

def is_valid_isbn13(s: str) -> bool:
    """Check for 13-digit-only form."""
    return is_valid_ean13(s)



# SPDX-License-Identifier: MIT
"""Known Vlju subtypes."""

from fnattr.vlju import Vlju
from fnattr.vlju.types.doi import DOI
from fnattr.vlju.types.ean import EAN13
from fnattr.vlju.types.ean.isbn import ISBN
from fnattr.vlju.types.ean.ismn import ISMN
from fnattr.vlju.types.ean.issn import ISSN
from fnattr.vlju.types.file import File
from fnattr.vlju.types.info import Info
from fnattr.vlju.types.lccn import LCCN
from fnattr.vlju.types.timestamp import Timestamp
from fnattr.vlju.types.uri import URI
from fnattr.vlju.types.url import URL
from fnattr.vlju.types.urn import URN

# yapf: disable
VLJU_TYPES: dict[str, type[Vlju]] = {
    'doi':      DOI,
    'ean':      EAN13,
    'file':     File,
    'info':     Info,
    'isbn':     ISBN,
    'ismn':     ISMN,
    'issn':     ISSN,
    'lccn':     LCCN,
    't':        Timestamp,
    'uri':      URI,
    'url':      URL,
    'urn':      URN,
}
# yapf: enable



# SPDX-License-Identifier: MIT
"""Encode and decode VljuMap."""

import csv as py_csv
import io
import json as py_json
import re
import shlex
import warnings

from collections.abc import Callable, Generator, Iterable
from typing import NamedTuple

from fnattr.util import escape
from fnattr.util.error import Error
from fnattr.util.multimap import MultiMap
from fnattr.vlju.types.ean.isbn import is_valid_isbn10, is_valid_isbn13
from fnattr.vljumap import VljuFactory, VljuMap

EncodeCallable = Callable[[VljuMap, str | None], str]
DecodeCallable = Callable[[VljuMap, str, VljuFactory], VljuMap]

class Encoder:
    """Scheme for for encoding and decoding VljuMap."""

    def __init__(self,
                 name: str,
                 encode: EncodeCallable | None,
                 decode: DecodeCallable | None,
                 desc: str | None = None,
                 description: str | None = None) -> None:
        self.name = name
        self.desc = desc or name
        self.description = description
        self.encode: EncodeCallable = encode or _unimplemented_encode
        self.decode: DecodeCallable = decode or _unimplemented_decode

    def can_encode(self) -> bool:
        return self.encode != _unimplemented_encode

    def can_decode(self) -> bool:
        return self.decode != _unimplemented_decode

def _unimplemented_decode(_n: VljuMap, _s: str,
                          _factory: VljuFactory) -> VljuMap:
    raise NotImplementedError

def _unimplemented_encode(_n: VljuMap, _mode: str | None = None) -> str:
    raise NotImplementedError

# Global registry of named Encoders.
encoder: dict[str, Encoder] = {}
decoder: dict[str, Encoder] = {}

def _register_encoder(e: Encoder) -> Encoder:
    if e.can_encode():  # pragma: no branch
        encoder[e.name] = e
    if e.can_decode():
        decoder[e.name] = e
    return e

###############################################################################
#
# V3 Encoder
#
###############################################################################

V3_DESC = 'Default.'

V3_GRAMMAR = """
    v3        → v3seq «‘ ’» v3title «‘ ’» v3attrs
    v3title   → [title [‘ - ’ title]*]
    v3attrs   → [‘[’ v3kv [‘; ’ v3kv]* ‘]’]
    v3kv      → k ‘=’ v
    v3seq     → [digit (alnum | ‘.’)* ‘.’]
    a «j» b   → (a | b | ajb)
"""

V3_DESCRIPTION = """
  Encoder format v3 consists, in order, of optional sequence numbers,
  optional title and subtitles, and optional attributes.

  Sequence numbers begin with a digit and end with a period. Multiple sequence
  numbers are allowed, but they must be adjacent.

  Title and optional subtitles are separated by ‘ - ’ (including the spaces).

  Attributes are surrounded by ‘[ … ]’ and separated by ‘;’. (A space follows
  each semicolon when encoding, but is not required when decoding.) Each
  attribute consists of a key, optionally followed by ‘=’ and a value.

  Characters with special meaning to the encoding, or not allowed in file
  names, are represented using URL-style % encoding.
""" + V3_GRAMMAR

class V3Config(NamedTuple):
    """Properties used by V3 and variants."""

    quote: escape.Escape
    attr_start: str = '['
    attr_end: str = ']'
    attr_join: str = '; '
    attr_kv: str = '='
    title_join: str = ' - '
    seq_start: str = ''
    seq_end: str = '.'
    seq_join: str = '.'

V3_CONFIG = V3Config(quote=escape.unixfile)

def v3_encode(n: VljuMap, mode: str | None = None) -> str:
    return _v3_enc(V3_CONFIG, n, mode)

def v3_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return _v3_dec(V3_CONFIG, n, s, factory)

v3 = _register_encoder(
    Encoder('v3', v3_encode, v3_decode, V3_DESC, V3_DESCRIPTION))

def _v3_enc(config: V3Config, n: VljuMap, mode: str | None) -> str:
    m = n.to_strings(mode)

    sequence = config.seq_join.join(m['n'])
    sequence = sequence and f'{config.seq_start}{sequence}{config.seq_end}'
    title_qjoin = config.title_join.translate(
        escape.mktrans_urlish(config.title_join.strip()))
    title = config.title_join.join(
        config.quote.encode(i).replace(config.title_join, title_qjoin)
        for i in m['title'])
    attrs = config.attr_join.join(
        kv_fmt(k, v, config.attr_kv, config.quote) for k,
        v in m.pairs() if k not in ('title', 'n'))
    attrs = attrs and f'{config.attr_start}{attrs}{config.attr_end}'
    return join_non_empty(' ', sequence, title, attrs)

def _v3_dec(config: V3Config, n: VljuMap, s: str,
            factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_v3_dec_iter(config, s), factory)

def _v3_dec_iter(config: V3Config, s: str) -> Iterable[tuple[str, str]]:
    if config.attr_start in s:
        s, _, attr = s.partition(config.attr_start)
    else:
        attr = ''
    if s:
        sequence, title = _v3_dec_seq(config, s)
        for v in sequence:
            yield ('n', v)
        if title:
            for i in title.split(config.title_join):
                yield ('title', escape.winfile.decode(i.strip()))
    if attr:
        for k, v in _v3_dec_attr(config, attr):
            yield (k, v)

def _v3_dec_attr(config: V3Config, s: str) -> Iterable[tuple[str, str]]:
    if s.endswith(config.attr_end):
        s = s[:-1]
    else:
        warnings.warn(f"Expected '{config.attr_end}' after '{s}'", stacklevel=0)
    for kv in s.split(config.attr_join.strip()):
        if config.attr_kv in kv:
            k, v = kv.split(config.attr_kv, 1)
        else:
            k = kv
            v = ''
        k = escape.winfile.decode(k.strip())
        v = escape.winfile.decode(v.strip())
        if not k:
            continue
        yield (k, v)

def _v3_dec_seq(config: V3Config, s: str) -> tuple[Iterable[str], str]:
    # Not currently used by any encoding:
    # if config.seq_start:
    #     if not s.startswith(config.seq_start):
    #         return ([], s)
    #     s = s[len(config.seq_start):]
    if config.seq_end not in s:
        return ([], s)
    seq: list[str] = []
    t = s
    lj = len(config.seq_join)
    while (((i := t.find(config.seq_join)) > 0) and t[: i].isalnum()
           and t[i + lj].isalnum() and t[0].isdigit()):
        seq.append(t[: i])
        t = t[(i + lj):]
    if ((i := t.find(config.seq_end)) > 0 and t[: i].isalnum()
            and t[0].isdigit()):
        seq.append(t[: i])
        t = t[(i + len(config.seq_end)):]
        return (seq, t.strip())
    return ([], s)

###############################################################################
#
# Windows Encoder
#
# Same as V3 except restricted to Windows file name characters.
#
###############################################################################

WIN_DESC = 'Like v3, but escape Windows forbidden characters.'
WIN_DESCRIPTION = """
  This is the same as v3 encoding format, except that additional characters
  are URL-escaped to comply with Windows file name limitations.
"""

WIN_CONFIG = V3Config(quote=escape.winfile)

def win_encode(n: VljuMap, mode: str | None = None) -> str:
    return _v3_enc(WIN_CONFIG, n, mode)

def win_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return _v3_dec(WIN_CONFIG, n, s, factory)

win = _register_encoder(
    Encoder('win', win_encode, win_decode, WIN_DESC, WIN_DESCRIPTION))

###############################################################################
#
# V2 Encoder
#
# Same as V3 except attributes are surrounded by ‘{}’ instead of ‘[]’
# and joined by ‘;’ with no space.
#
###############################################################################

V2_DESC = 'An obsolete encoder.'

V2_GRAMMAR = """
    v2        → v2seq «‘ ’» v2title «‘ ’» v2attrs
    v2title   → [title [‘ - ’ title]*]
    v2attrs   → [‘{’ v2kv [‘;’ v2kv]* ‘}’]
    v2kv      → k ‘=’ v
    v2seq     → [digit (alnum | ‘.’)* ‘.’]
    a «j» b   → (a | b | ajb)
"""

V2_DESCRIPTION = """
  Encoder format v2 is similar to v3, except that attributes are surrounded
  by ‘{ … }' rather than ‘[ … ]’, and separated by semicolons with no space.

  This is supported only to covert old file names.
""" + V2_GRAMMAR

V2_CONFIG = V3Config(
    quote=escape.unixfile, attr_start='{', attr_join=';', attr_end='}')

def v2_encode(n: VljuMap, mode: str | None = None) -> str:
    return _v3_enc(V2_CONFIG, n, mode)

def v2_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return _v3_dec(V2_CONFIG, n, s, factory)

v2 = _register_encoder(
    Encoder('v2', v2_encode, v2_decode, V2_DESC, V2_DESCRIPTION))

###############################################################################
#
# V1 Encoder
#
###############################################################################

V1_DESC = 'An obsolete encoder.'

V1_GRAMMAR = """
    v1        → v1author ‘:’ «‘ ’» v1title «‘ ’» v1attrs
    v1author  → [author [‘; ’ author]*]
    v1title   → [title [‘: ’ title]*]
    v1attrs   → [‘[’ v1kv [‘,’ v1kv]* ‘]’]
    v1kv      → k ‘=’ v
    a «j» b   → (a | b | ajb)
"""

V1_DESCRIPTION = """
  Encoder format v1 begins with optional authors, followed by optional
  title and subtitle, followed by optional attributes.

  Authors are separated by ‘; ’ and followed by a ‘:’.

  Titles and subtitles are separated by ‘: ’.

  Attributes are surrounded by ‘[ … ]’ and separated by ‘,’.

  This is supported only to covert old file names; it should not be used
  due to the ambiguity between authors and titles with subtitles.
""" + V1_GRAMMAR

V1_CONFIG = V3Config(
    escape.unixfile, attr_start='[', attr_end=']', attr_join=',')

def v1_encode(n: VljuMap, mode: str | None = None) -> str:
    m = n.to_strings(mode)
    r = _v1_enc_author_title(m)
    attrs = ','.join(
        kv_fmt(k, v, '=', escape.unixfile) for k,
        v in m.pairs() if k not in ('title', 'a'))
    return spj(r, f'[{attrs}]') if attrs else r

def v1_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_v1_dec_iter(V1_CONFIG, s), factory)

v1 = _register_encoder(
    Encoder('v1', v1_encode, v1_decode, V1_DESC, V1_DESCRIPTION))

def _v1_enc_author_title(m: MultiMap) -> str:
    author = '; '.join(i for i in m['a'])
    r = f'{author}:' if author else ''
    title = ': '.join(i for i in m['title'])
    return spj(r, title)

def _v1_dec_iter(config: V3Config, s: str) -> Iterable[tuple[str, str]]:
    if '[' in s:
        s, a = s.split('[', 1)
    else:
        a = ''
    for k, v in _v1_dec_author_title(s):
        yield (k, v)
    if a:
        for k, v in _v3_dec_attr(config, a):
            yield (k, v)

def _v1_dec_author_title(s: str) -> Generator[tuple[str, str], None, None]:
    if ':' in s:
        author, s = s.split(':', 1)
        for i in author.split(';'):
            yield ('a', escape.winfile.decode(i.strip()))
    for i in s.split(':'):
        if i:
            yield ('title', escape.winfile.decode(i.strip()))

###############################################################################
#
# V0 Encoder
#
###############################################################################

V0_DESC = 'The most obsolete encoder.'

V0_GRAMMAR = """
    v0        → v1author «‘ ’» v1title «‘ ’» (isbn | ‘lccn=’ lccn)
    a «j» b   → (a | b | ajb)
"""

V0_DESCRIPTION = """
  Encoder format v0 is strictly limited in the attributes it can represent.
  This is supported only to convert old file names.
""" + V0_GRAMMAR

def v0_encode(n: VljuMap, mode: str | None = None) -> str:
    m = n.to_strings(mode)
    r = _v1_enc_author_title(m)
    if 'isbn' in m:
        return spj(r, m['isbn'][0])
    if 'lccn' in m:
        return spj(r, f'lccn={m["lccn"][0]}')
    return r

def v0_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_v0_dec_iter(s), factory)

v0 = _register_encoder(
    Encoder('v0', v0_encode, v0_decode, V0_DESC, V0_DESCRIPTION))

V0_RE = re.compile(
    r"""
        (?P<rest>.*)
        \b(?:
            (?P<isbn> (?: [0-9]{13} | [0-9]{9}[0-9xX] ) )
        | lccn=(?P<lccn> \S+ )
        )$
        """,
    re.VERBOSE)

def _v0_dec_iter(s: str) -> Generator[tuple[str, str], None, None]:
    if m := V0_RE.fullmatch(s):
        s = m.group('rest')
    yield from _v1_dec_author_title(s)
    if m:
        if isbn := m.group('isbn'):
            yield ('isbn', isbn)
        elif lccn := m.group('lccn'):   # pragma: no branch
            yield ('lccn', lccn)
        else:                           # pragma: no cover
            message = f'no isbn or lccn in {s}'
            raise Error(message)

###############################################################################
#
# SFC Encoder
#
###############################################################################

SFC_DESC = "SFC's encoder."

SFC_GRAMMAR = """
    sfc       → sfctitle [‘ by ’ sfcauthor] [«‘ ’» (isbn | date | sfced)]*
    sfctitle  → [title [‘ - ’ title]*]
    sfcauthor → author [‘, ’ author]*
    sfced     → edition (‘st’ | ‘nd’ | ‘rd’ | ‘th’) ‘ edition’
    a «j» b   → (a | b | ajb)
"""

SFC_DESCRIPTION = """
  Encoder format sfc consists of a title and optional subtitles, optional
  authors, and optional specific attributes: ISBN, year, and edition.

  Title and optional subtitles are separated by ‘ - ’ (including the spaces).

  Authors are preceded by ‘ by ’ and separated by commas.

  An ISBN may follow. A four-digit year may follow. An edition, consisting
  of a number, a number suffix, and the word ‘edition’, may follow.
""" + SFC_GRAMMAR

def sfc_encode(n: VljuMap, mode: str | None = None) -> str:
    m = n.to_strings(mode)
    title = ' - '.join(i for i in m['title'])
    author = ', '.join(i for i in m['a'])
    if author:
        author = f'by {author}'
    isbn = m['isbn'][0] if 'isbn' in m else ''
    if 'edition' in m:
        e = int(m['edition'][0])
        edition = f'{e}{nth(e)} edition'
    else:
        edition = ''
    date = m['date'][0] if 'date' in m else ''
    return join_non_empty(' ', title, author, isbn, edition, date)

def sfc_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_sfc_dec_iter(s), factory)

sfc = _register_encoder(
    Encoder('sfc', sfc_encode, sfc_decode, SFC_DESC, SFC_DESCRIPTION))

SFC_TAIL_RE = re.compile(
    r"""
        (?P<prefix> .*)
        (?:
            (?P<date> [12]\d\d\d ) |
            (?: (?P<edition> \d+ ) \w*\s+edition )
        )$
        """,
    re.X)

def _sfc_dec_iter(s: str) -> Generator[tuple[str, str], None, None]:
    while True:
        if m := SFC_TAIL_RE.fullmatch(s):
            for k in ('date', 'edition'):
                if v := m.group(k):
                    yield (k, v)
            s = m.group('prefix').strip()
        elif is_valid_isbn10(s[-10 :]):
            yield ('isbn', s[-10 :])
            s = s[:-10].strip()
        elif is_valid_isbn13(s[-13 :]):
            yield ('isbn', s[-13 :])
            s = s[:-13].strip()
        else:
            break
    if (by := s.rfind(' by ')) > 0:
        authors = s[by + 4 :]
    elif s.startswith('by '):
        authors = s[3 :]
        by = 0
    else:
        authors = None
    if authors:
        s = s[: by].strip()
        a = list(map(str.strip, authors.split(',')))
        while a:
            if len(a) > 1 and ' ' not in a[0] and ' ' not in a[1]:
                yield ('a', f'{a[0]}, {a[1]}')
                a = a[2 :]
            else:
                yield ('a', a[0])
                a = a[1 :]
    if s:
        for i in s.split(' - ', 1):
            yield ('title', escape.winfile.decode(i.strip()))

###############################################################################
#
# JSON
#
###############################################################################

JSON_DESC = 'Javascript Object Notation'

JSON_DESCRIPTION = """
  Attributes are encoded in JSON, as an object where each key contains a list
  of values.
"""

def json_encode(n: VljuMap, mode: str | None = None) -> str:
    return py_json.dumps(dict(n.get_lists(mode)))

def json_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_json_dec_iter(s), factory)

def _json_dec_iter(s: str) -> Generator[tuple[str, str], None, None]:
    for k, vl in py_json.loads(s).items():
        if isinstance(vl, list):
            for v in vl:
                yield (k, str(v))
        else:
            yield (k, str(vl))

json = _register_encoder(
    Encoder('json', json_encode, json_decode, JSON_DESC, JSON_DESCRIPTION))

###############################################################################
#
# Shell
#
###############################################################################

SHELL_DESC = 'Shell arrays'

SHELL_DESCRIPTION = """
  Attributes are encoded as shell arrays (ksh, bash).
  Decoding is not implemented.
"""

def shell_encode(n: VljuMap, mode: str | None = None) -> str:
    r = []
    for k, vlist in n.get_lists(mode):
        v = ' '.join(shlex.quote(v) for v in vlist if v is not None)
        r.append(f'{k}=({v})')
    return '\n'.join(r)

sh = _register_encoder(
    Encoder('sh',
            shell_encode,
            _unimplemented_decode,
            SHELL_DESC,
            SHELL_DESCRIPTION))

###############################################################################
#
# value
#
###############################################################################

VALUE_DESC = 'Bare values'

VALUE_DESCRIPTION = """
  Attributes are presented as raw strings, one per line.
  Since keys are not represented, decoding is not possible.
"""

def value_encode(n: VljuMap, mode: str | None = None) -> str:
    return '\n'.join((f'{v or k}' for k, v in n.get_pairs(mode)))

value = _register_encoder(
    Encoder('value',
            value_encode,
            _unimplemented_decode,
            VALUE_DESC,
            VALUE_DESCRIPTION))

###############################################################################
#
# key/value
#
###############################################################################

KEYVALUE_DESC = 'Key:Value pairs'

KEYVALUE_DESCRIPTION = """
  Attributes are presented as a key, followed by ‘: ’, follow by a value.
  Each attribute (including multiple attributes for the same key) appears
  on its own line.
"""

def keyvalue_encode(n: VljuMap, mode: str | None = None) -> str:
    return '\n'.join((f'{k}: {v}' for k, v in n.get_pairs(mode)))

def keyvalue_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_keyvalue_dec_iter(s), factory)

def _keyvalue_dec_iter(s: str) -> Generator[tuple[str, str], None, None]:
    for kv in s.split('\n'):
        if kv:
            k, v = kv.split(':', 1)
            yield (k.strip(), v.strip())

keyvalue = _register_encoder(
    Encoder('keyvalue',
            keyvalue_encode,
            keyvalue_decode,
            KEYVALUE_DESC,
            KEYVALUE_DESCRIPTION))

###############################################################################
#
# csv
#
###############################################################################

CSV_DESC = 'Comma-separated values'

CSV_DESCRIPTION = """
  Attributes are presented as a CSV table of two columns, the first being
  the key and the second the associated value.
"""

def csv_encode(n: VljuMap, mode: str | None = None) -> str:
    return _csv_enc(n, mode, dialect='unix')

def csv_decode(n: VljuMap, s: str, factory: VljuFactory) -> VljuMap:
    return n.add_pairs(_csv_dec_iter(s, dialect='unix'), factory)

csv = _register_encoder(
    Encoder('csv', csv_encode, csv_decode, CSV_DESC, CSV_DESCRIPTION))

def _csv_enc(n: VljuMap, mode: str | None, **kwargs) -> str:
    with io.StringIO() as f:
        w = py_csv.writer(f, **kwargs)
        for kv in n.get_pairs(mode):
            w.writerow(kv)
        return f.getvalue()

def _csv_dec_iter(s: str, **kwargs) -> Generator[tuple[str, str], None, None]:
    for row in py_csv.reader(s.split('\n'), **kwargs):
        if row:
            k, v = row
            yield (k, v)

###############################################################################
#
# Utility functions
#
###############################################################################

def kv_fmt(k: str, v: str | None, sep: str, e: escape.Escape) -> str:
    if v:
        return f'{k}{sep}{e.encode(v)}'
    return k

def join_non_empty(sep: str, *args: str) -> str:
    """Join non-empty args with the given separator."""
    return sep.join(filter(bool, args))

def spj(s: str, t: str) -> str:
    return join_non_empty(' ', s, t)

def nth(n: int | str) -> str:
    """Return English suffix for ordinal numbers."""
    n = int(n)
    return ('th', 'st', 'nd', 'rd', 'th', 'th', 'th', 'th', 'th', 'th')[n % 10]



# SPDX-License-Identifier: MIT
"""Argument polymorphism utility."""

from collections.abc import Iterator, KeysView, Mapping
from typing import Generic, Self, TypeVar

T = TypeVar('T')

class Registry(Generic[T]):
    """
    Used to enable argument polymorphism.

    Several operations have arguments that can take any of three forms:
    - a value of a type T;
    - a string, looked up to get a value of type T;
    - None, indicating a default value.
    """

    def __init__(self, name: str | None = None) -> None:
        self.dict: dict[str, T] = {}
        self.default: T | None = None
        self.name = name

    def __copy__(self) -> Self:
        r = type(self)(self.name)
        r.dict = self.dict.copy()
        r.default = self.default
        return r

    def __iter__(self) -> Iterator[str]:
        return iter(self.dict)

    def update(self, m: Mapping[str, T]) -> Self:
        self.dict.update(m)
        return self

    def set_default(self, e: T | str | None) -> Self:
        self.default = self.get(e)
        return self

    def get(self, t: T | str | None = None) -> T:
        if t is None:
            if self.default is None:
                msg = 'no default value'
                raise self._keyerror(msg)
            return self.default
        if isinstance(t, str):
            try:
                return self.dict[t]
            except KeyError as e:
                msg = f'‘{t}’ is not known'
                raise self._keyerror(msg) from e
        return t

    def keys(self) -> KeysView[str]:
        return self.dict.keys()

    def _keyerror(self, message: str) -> KeyError:
        if self.name:
            message = self.name + ': ' + message
        return KeyError(message)



# SPDX-License-Identifier: MIT
"""URN - Vlju representable as a URN."""

from fnattr.util.repr import mkrepr
from fnattr.vlju.types.uri import URI, Authority

class URN(URI):
    """
    Represents a URN.

    short:  uri
    long:   uri
    """

    def __init__(self,
                 s: str,
                 authority: Authority | str | None = None,
                 q: str | None = None,
                 r: str | None = None) -> None:
        super().__init__(
            s,
            scheme='urn',
            authority=authority,
            urnq=q,
            urnr=r,
            sa=':',
            ap=':')

    # URI overrides:

    def sauthority(self) -> str:
        a = self.authority()
        return str(a.host) if a else ''

    # Vlju overrides:

    def __eq__(self, other: object) -> bool:
        if isinstance(other, URN):
            return (self._value == other._value                 # noqa: SLF001
                    and self._scheme == other._scheme           # noqa: SLF001
                    and self._authority == other._authority     # noqa: SLF001
                    and self._urnq == other._urnq               # noqa: SLF001
                    and self._urnr == other._urnr)              # noqa: SLF001
        return False

    def __repr__(self) -> str:
        return mkrepr(self, ['_value'], ['_authority', '_urnq', '_urnr'])



# SPDX-License-Identifier: MIT
"""Pre-configured VljuM."""

from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any

from fnattr.util.registry import Registry
from fnattr.vlju.types.all import VLJU_TYPES, Vlju
from fnattr.vlju.types.site import site_class
from fnattr.vljum import VljuM
from fnattr.vljumap import enc
from fnattr.vljumap.factory import (
    LooseMappedFactory,
    MappedFactory,
    default_factory,
)

V = Vlju

class M(VljuM):
    """Configured subclass of VljuM."""

    raw_factory = default_factory
    strict_factory = MappedFactory(VLJU_TYPES)
    loose_factory = LooseMappedFactory(VLJU_TYPES)
    default_registry = {
        'factory':
            Registry().update({
                'raw': raw_factory,
                'typed': loose_factory,
                'loose': loose_factory,
                'strict': strict_factory,
            }).set_default('loose'),
        'encoder':
            Registry().update(enc.encoder).set_default('v3'),
        'decoder':
            Registry().update(enc.decoder).set_default('v3'),
        'mode':
            Registry().update({
                k: k
                for k in ('short', 'long', 'repr')
            }).set_default('short'),
    }

    @classmethod
    def configure_sites(cls, site: Mapping[str, Mapping[str, Any]]) -> None:
        for k, s in site.items():
            scls = site_class(**s)
            cls.strict_factory.setitem(k, scls)
            cls.loose_factory.setitem(k, scls)

    @classmethod
    def exports(cls) -> dict[str, Any]:
        x = EXPORTS.copy()
        for k, v in cls.strict_factory.kmap.items():
            x[k] = v
            x[v.__name__] = v
        return x

    @classmethod
    def evaluate(cls,
                 s: str,
                 g: dict[str, Any] | None = None) -> Any:   # noqa: any-type
        if g is None:
            g = cls.exports()
        return eval(s, g)                                   # noqa: eval

    @classmethod
    def execute(cls, s: str, g: dict[str, Any] | None = None) -> dict[str, Any]:
        if g is None:
            g = cls.exports()
        exec(s, g)  # noqa: exec-builtin
        return g

def _make_free_function(cls: type, name: str) -> Callable:
    method = getattr(cls, name)

    def f(*args, **kwargs) -> Any:  # noqa: ANN401
        return method(cls(), *args, **kwargs)

    return f

EXPORTS: dict[str, Any] = {
                                                            # Aliases
    'Path': Path,
    'V': Vlju,
} | {
                                                            # Module definitions
    k: globals()[k]
    for k in ('M', )
} | {
                                                            # M() methods
    k: _make_free_function(M, k)
    for k in ('add', 'decode', 'file', 'read', 'reset')
}



# SPDX-License-Identifier: MIT
"""Utilities for testing."""

from collections.abc import Iterable, Iterator, Mapping, Sequence

def im2p(cases: Iterable[Mapping],
         keys: Sequence[str] | None = None) -> tuple[str, list]:
    """Convert Mappings to pytest Parameters."""
    if keys is None:
        keys = list(next(iter(cases)).keys())
    return _2p(iter(cases), keys, keys)

def it2p(cases: Iterable[tuple],
         keys: Sequence[str] | None = None) -> tuple[str, list]:
    """Convert tuples to pytest Parameters. First entry is field names."""
    it = iter(cases)
    names = next(it)
    if keys is None:
        keys = names
    return _2p(it, keys, [names.index(k) for k in keys])

def _2p(it: Iterator, keys: Sequence[str],
        indices: Sequence) -> tuple[str, list]:
    if len(indices) == 1:
        # For this case, pytest requires bare items, not single-element tuples.
        index = indices[0]
        r = [i[index] for i in it]
    else:
        r = [tuple(i[index] for index in indices) for i in it]
    return (','.join(keys), r)



# SPDX-License-Identifier: MIT
"""Command DSL."""

import textwrap

from collections.abc import Callable, Iterable, Iterator

from fnattr.util.docsplit import docsplit
from fnattr.util.error import Error
from fnattr.util.registry import Registry
from fnattr.vljum.m import M

class Runner:
    """Command DSL."""

    commands: dict[str, Callable] = {}

    def __init__(self, m: M | None = None) -> None:
        self.tokens: Iterator[str] | None = None
        self.m = M() if m is None else m
        self.report = False
        self.help: dict | None = None
        self.commands = {}
        for i in dir(type(self)):
            if i.startswith('command_'):
                m = getattr(type(self), i)
                if callable(m):             # pragma: no branch
                    self.commands[i[8 :]] = m
        self.commands |= {
            # Factories
            k: type(self).set_factory
            for k in self.m.factory
        } | {
            # Encoders/decoders.
            k: type(self).set_coder
            for k in self.m.encoder
        } | {
            # Modes.
            k: type(self).set_mode
            for k in self.m.mode
        }

    def token(self) -> str | None:
        if self.tokens:
            try:
                return next(self.tokens)
            except StopIteration:
                pass
        return None

    def need(self, message: str = 'expected token') -> str:
        t = self.token()
        if t is None:
            raise Error(message)
        return t

    def command_add(self, cmd: str) -> None:
        """
        Add an attribute.

        Constructs the value using the current active factory.

        Synopsis: add ‹key› ‹value›
        """
        key = self.need(f'{cmd}: expected key')
        val = self.need(f'{cmd}: expected value')
        self.m.add(key, val)
        self.report = True

    def command_compare(self, _: str) -> None:
        """
        Print the original and current encoded file name if they differ.

        Encodes using the current active encoder.
        This can be used as a ‘dry run’ for `rename`.

        Synopsis: compare
        """
        original = self.m.original()
        current = self.m.filename()
        if current != original:
            print(original)
            print(current)
        self.report = False

    def command_decode(self, cmd: str) -> None:
        """
        Decode a string.

        Decodes using the current active decoder.

        Synopsis: decode ‹string›
        """
        self.m.decode(self.need(f'{cmd}: expected string'))
        self.report = True

    def command_decoder(self, cmd: str) -> None:
        """
        Set the current active decoder.

        Synopsis: decoder ‹decoder›
        """
        self.set_registry(self.m.decoder, cmd)

    def command_delete(self, cmd: str) -> None:
        """
        Delete all attributes for one or more ‹key›s.

        It is not an error for keys not to be present.
        The complement of `delete` is `extract`.

        Synopsis: delete ‹key›[,‹key›]*
        """
        for key in self.need(f'{cmd}: expected keys').split(','):
            self.m.remove(key)
        self.report = True

    def command_dir(self, cmd: str) -> None:
        """
        Set the directory associated with a file name.

        Synopsis: dir ‹directory›
        """
        self.m.with_dir(self.need(f'{cmd}: expected directory'))
        self.report = True

    def command_encode(self, _: str) -> None:
        """
        Encode and prints the current attributes.

        Encodes using the current active encoder.

        Synopsis: encode
        """
        print(self.m.encode())
        self.report = False

    def command_encoder(self, cmd: str) -> None:
        """
        Set the current active encoder.

        Synopsis: encoder ‹encoder›
        """
        self.set_registry(self.m.encoder, cmd)

    def command_extract(self, cmd: str) -> None:
        """
        Extract attributes for one or more keys.

        It is not an error for keys not to be present.
        The complement of `extract` is `delete`.

        Synopsis: extract ‹key›[,‹key›]*
        """
        self.m = self.m.submap(self.need(f'{cmd}: expected keys').split(','))
        self.report = True

    def command_factory(self, cmd: str) -> None:
        """
        Set the current active factory.

        Synopsis: factory ‹factory›
        """
        self.set_registry(self.m.factory, cmd)

    def command_file(self, cmd: str) -> None:
        """
        Decode a file name.

        Decodes using the current active decoder.

        Synopsis: file ‹filename›
        """
        self.m.file(self.need(f'{cmd}: expected filename'))
        self.report = True

    def command_filename(self, _: str) -> None:
        """
        Encode and print the current attributes as a file name.

        Encodes using the current active encoder.

        Synopsis: encode
        """
        print(self.m.filename())
        self.report = False

    def command_help(self, _: str) -> None:
        """
        Show information about a subcommand, or list subcommands.

        Synopsis: help [‹subcommand›]*
        """
        if self.help is None:   # pragma: no branch
            self.help = {}
            for k, c in self.commands.items():
                if c.__doc__:
                    self.help[k] = docsplit(c.__doc__)
        helped = 0
        while (name := self.token()) is not None:
            helped += 1
            if name not in self.help:
                print(f'No help available for {name}\n')
                helped = 0
                break
            desc, info = self.help[name]
            print(f'NAME\n  {name} - {desc[0]}\n')
            synopsis = info.get('synopsis')
            if synopsis:        # pragma: no branch
                print(f'SYNOPSIS\n  {synopsis}\n')
            description = '\n\n'.join(desc)
            if description:     # pragma: no branch
                print(f'DESCRIPTION\n{textwrap.indent(description, "  ")}\n')
        if helped == 0:
            print('COMMANDS')
            for name in sorted(self.help.keys()):
                print(f'  {name:8} - {self.help[name][0][0]}')

    def command_mode(self, cmd: str) -> None:
        """
        Set the current active mode.

        Synopsis: mode ‹mode›
        """
        self.set_registry(self.m.mode, cmd)

    def command_order(self, cmd: str) -> None:
        """
        Arranges keys.

        With `--all`, arranges the attribute keys in alphabetical order.

        With given ‹key›s, arranges the attribute set so that those keys appear
        in the specified order. Other keys will follow in their original order.

        Synopsis: order (--all | ‹key›[,‹key›]*)
        """
        t = self.need(f'{cmd}: expected keys')
        self.m = self.m.sortkeys(None if t == '--all' else t.split(','))
        self.report = True

    def command_quiet(self, _: str) -> None:
        self.report = False

    def command_remove(self, cmd: str) -> None:
        """
        Remove a specific attribute.

        Constructs the value using the current active factory.

        Synopsis: remove ‹key› ‹value›
        """
        key = self.need(f'{cmd}: expected key')
        val = self.need(f'{cmd}: expected value')
        self.m.remove(key, val)
        self.report = True

    def command_rename(self, _: str) -> None:
        """
        Rename a file.

        If attributes, directory, or suffix have changed since the
        file name decoding, rename the original file according
        to the current state.

        Synopsis: rename
        """
        self.m.rename()
        self.report = False

    def command_set(self, cmd: str) -> None:
        """
        Set an attribute.

        Replaces any existing attributes for the same ‹key›.
        Constructs the value using the current active factory.

        Synopsis: set ‹key› ‹value›
        """
        key = self.need(f'{cmd}: expected key')
        val = self.need(f'{cmd}: expected value')
        self.m.reset(key, val)
        self.report = True

    def command_sort(self, cmd: str) -> None:
        """
        Sorts values for a given key or all keys.

        Synopsis: sort (--all | ‹key›[,‹key›]*)
        """
        t = self.need(f'{cmd}: expected keys or ‘--all’')
        if t == '--all':
            self.m.sort()
        else:
            self.m.sort(*t.split(','))
        self.report = True

    def command_suffix(self, cmd: str) -> None:
        """
        Set the suffix associated with a file name.

        Synopsis: suffix ‹ext›
        """
        self.m.with_suffix(self.need(f'{cmd}: expected suffix'))
        self.report = True

    def command_uri(self, _: str) -> None:
        """
        Print attribute URI(s).

        If multiple attributes have URIs, they will be printed on separate
        lines.

        Synopsis: uri
        """
        print(self.m.uri())
        self.report = False

    def command_url(self, _: str) -> None:
        """
        Print attribute URL(s).

        If multiple attributes have URLs, they will be printed on separate
        lines.

        Synopsis: url
        """
        u = self.m.url()
        if u:
            print(u)
        self.report = False

    def set_coder(self, cmd: str) -> None:
        if cmd in self.m.decoder:
            self.m.decoder.set_default(cmd)
        if cmd in self.m.encoder:   # pragma: no branch
            self.m.encoder.set_default(cmd)

    def set_factory(self, cmd: str) -> None:
        self.m.factory.set_default(cmd)

    def set_mode(self, cmd: str) -> None:
        self.m.mode.set_default(cmd)

    def set_registry(self, r: Registry, cmd: str) -> None:
        keys = r.keys()
        choices = ', '.join(f'‘{k}’' for k in keys)
        message = f'{cmd}: expected one of: {choices}'
        t = self.need(message)
        if t not in keys:
            raise Error(message)
        r.set_default(t)

    def run(self, args: Iterable[str]) -> None:
        self.tokens = iter(args)
        self.report = False
        while (cmd := self.token()) is not None:
            c = self.commands.get(cmd)
            if not c:
                msg = f'{cmd}: Unknown command'
                raise Error(msg)
            c(self, cmd)
        if self.report:
            print(self.m)

    def runs(self, s: str) -> None:
        self.run(s.split())



# SPDX-License-Identifier: MIT
"""ISMN (International Standard Music Number)."""

from fnattr.vlju.types.ean import EAN13, as13

class ISMN(EAN13):
    """Represents an ISMN (International Standard Music Number)."""

    def __init__(self, s: str) -> None:
        v = as13(s, 'ismn')
        if v is None:
            msg = f'value {s} is not an ISMN'
            raise ValueError(msg)
        super().__init__(v, 'ismn')

    def lv(self) -> str:
        if self._value[: 4] != '9790':      # pragma: no branch
            raise ValueError(self._value)   # pragma: no cover
        return f'M{self._value[4:]}'

    def path(self) -> str:
        return self.lv()



# SPDX-License-Identifier: MIT
"""Move files to directories based on matching attributes."""

import argparse
import sys

from collections.abc import Iterable
from pathlib import Path

from fnattr.util.config import read_cmd_configs, xdg_config
from fnattr.vljum.m import M
from fnattr.vljumap import enc

def find_map_files(cmd: str) -> list[Path]:
    return [
        f for f in (xdg_config('vlju/fnaffle.map'),
                    xdg_config(f'{cmd}/fnaffle.map'),
                    Path('fnaffle.map'),
                    Path('.fnaffle')) if f is not None and f.exists()
    ]

Maps = list[tuple[Path, str, dict]]

def read_map_files(files: Iterable[Path | str]) -> Maps:
    maps: Maps = []
    for file in files:
        maps += read_map_file(file)
    return maps

def read_map_file(file: Path | str) -> Maps:
    maps = []
    with Path(file).open(encoding='utf-8') as f:
        while line := f.readline():
            dst, op, encoding = line.strip().split(None, 2)
            m = M().decode(encoding, 'v3')
            target = sorted_strings(m)
            maps.append((Path(dst), op, target))
    return maps

StringListDict = dict[str, list[str]]

def sorted_strings(m: M) -> StringListDict:
    d = {}
    for k in sorted(m.keys()):
        d[k] = sorted(str(v) for v in m[k])
    return d

def compare_sorted_lists(a: list, b: list) -> str:
    r = 3
    an = len(a)
    bn = len(b)
    ai = 0
    bi = 0
    while ai < an and bi < bn and r:
        if a[ai] < b[bi]:
            r &= ~1
            ai += 1
            continue
        if a[ai] > b[bi]:
            r &= ~2
            bi += 1
            continue
        ai += 1
        bi += 1
    if (r & 1) and ai < an:
        r &= ~1
    if (r & 2) and bi < bn:
        r &= ~2
    return ['≠', '⊂', '⊃', '='][r]

def match_map(op: str, source: StringListDict, target: StringListDict) -> bool:
    ok = {
        '⊆': ('⊂', '='),
        '⊇': ('⊃', '='),
    }.get(op, (op, ))
    for k, v in target.items():
        if not (s := source.get(k)):
            return False
        r = compare_sorted_lists(s, v)
        if r not in ok:
            return False
    return True

def match_maps(maps: Maps, m: M) -> Path | None:
    source = sorted_strings(m)
    for dst, op, target in maps:
        if match_map(op, source, target):
            return dst
    return None

def main(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv
    cmd = Path(argv[0]).stem
    parser = argparse.ArgumentParser(
        prog=cmd, description='Move files according to file name attributes')
    parser.add_argument(
        '--config',
        '-c',
        metavar='FILE',
        type=str,
        action='append',
        help='Configuration file.')
    parser.add_argument(
        '--decoder',
        '-d',
        metavar='DECODER',
        type=str,
        choices=enc.decoder.keys(),
        help='Default string decoder.')
    parser.add_argument(
        '--dryrun',
        '-n',
        default=False,
        action='store_true',
        help='Do not actually rename.')
    parser.add_argument(
        '--map',
        '-m',
        metavar='FILE',
        type=str,
        action='append',
        help='Renaming map file.')
    parser.add_argument(
        '--verbose',
        '-v',
        default=False,
        action='store_true',
        help='Report renaming.')
    parser.add_argument(
        'file',
        metavar='FILENAME',
        type=str,
        nargs=argparse.REMAINDER,
        default=[],
        help='File name(s).')
    args = parser.parse_args(argv[1 :])

    config = read_cmd_configs(cmd, args.config)
    options = config.get('option', {})
    if args.decoder:
        options['decoder'] = args.decoder
    M.configure_options(options)
    M.configure_sites(config.get('site', {}))

    try:
        if not args.map:
            args.map = find_map_files(cmd)
        if not args.map:
            print(f'{cmd}: no map file.')
            return 1
        maps = read_map_files(args.map)
        for file in args.file:
            m = M().file(file)
            if dst := match_maps(maps, m):
                if not dst.exists():
                    if not args.dryrun:
                        dst.mkdir(parents=True)
                    if args.dryrun or args.verbose:
                        print(f'{dst}: created')
                if not args.dryrun:
                    try:
                        m.with_dir(dst).rename()
                    except FileExistsError as e:
                        print(f'{cmd}: file exists: {e}')
                if args.dryrun or args.verbose:
                    print(f'{dst}: {file}')
            elif args.verbose:
                print(f'no match for {file}')
    except Exception as e:
        print(f'{cmd}: Unhandled exception: {type(e).__name__}{e.args}')
        raise
    return 0

if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: MIT
"""Helper for testing construction using cast_params()."""

from typing import Any

class CastParams:
    """Helper for testing construction using cast_params()."""

    def __init__(self, s: str | None, d: dict[str, Any]) -> None:
        self.s = s
        self.d = d

    def cast_params(self, _) -> tuple[str | None, dict[str, Any]]:
        return (self.s, self.d)



# SPDX-License-Identifier: MIT
"""LCCN - Library of Congress Control Number."""

# http://info-uri.info/registry/OAIHandler?verb=GetRecord&metadataPrefix=reg&identifier=info:lccn/

from fnattr.vlju.types.info import Info
from fnattr.vlju.types.uri import Authority
from fnattr.vlju.types.url import URL

def normalize(s: str) -> str:
    """Normalize according to info:lccn."""
    s = s.replace(' ', '')
    if (n := s.find('/')) > 0:
        s = s[: n]
    if (n := s.find('-')) > 0:
        s = f'{s[:n]}{int(s[n + 1:]):06}'
    return s.lower()

class LCCN(Info):
    """
    Represents an LCCN (Library of Congress Control Number).

    short:  value
    long:   uri
    """

    _lc = Authority('lccn.loc.gov')

    def __init__(self, s: str) -> None:
        super().__init__(normalize(s), authority='lccn')

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URL:
            return (self.spath(),
                    {
                        'scheme': 'https',
                        'authority': self._lc,
                        'query': self.query(),
                        'fragment': self.fragment(),
                    })
        return super().cast_params(t)

    # Vlju overrides

    def __str__(self) -> str:
        return self._value



# SPDX-License-Identifier: MIT
"""MultiMap."""

from collections import defaultdict
from collections.abc import (
    Callable,
    Generator,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
)
from typing import Generic, Self, TypeVar

K = TypeVar('K')
V = TypeVar('V')

T = TypeVar('T')
W = TypeVar('W')

class MultiMap(Generic[K, V]):
    """Multi-valued dictionary."""

    def __init__(self) -> None:
        self.data: defaultdict[K, list[V]] = defaultdict(list[V])

    def __eq__(self, other: object) -> bool:
        # pylint:disable=unidiomatic-typecheck
        if isinstance(other, MultiMap):
            return (type(self) == type(other)) and (self.data == other.data)
        return False

    def __getitem__(self, k: K) -> list[V]:
        return self.data[k]

    def __delitem__(self, k: K) -> None:
        if k in self.data:
            del self.data[k]

    def __contains__(self, k: K) -> bool:
        return (k in self.data) and (self.data[k] != [])

    def __len__(self) -> int:
        return len(self.data)

    def __repr__(self) -> str:
        return f'MultiMap({dict(self.data)!r})'

    def __iter__(self) -> Iterator[K]:
        return iter(self.data)

    def copy(self) -> Self:
        r = type(self)()
        r.data = self.data.copy()
        return r

    def get(self, k: K, default: list[V] | None = None) -> list[V]:
        if k in self.data:
            return self.data[k]
        if default is None:
            return []
        return default

    def keys(self) -> KeysView[K]:
        return self.data.keys()

    def pairs(self) -> Generator[tuple[K, V], None, None]:
        """Yield (k, v), potentially repeating k."""
        for k, vlist in self.data.items():
            for v in vlist:
                yield (k, v)

    def lists(self) -> ItemsView[K, list[V]]:
        """Yield (k, vlist)."""
        return self.data.items()

    def add(self, k: K, v: V) -> Self:
        if v not in self.data[k]:
            self.data[k].append(v)
        return self

    def remove(self, k: K, v: V) -> Self:
        self.data[k].remove(v)
        return self

    def pop(self, key: K, default: T | None = None) -> V | T | None:
        if self.data[key]:
            return self.data[key].pop()
        return default

    def top(self, k: K) -> V | None:
        if self.data[k]:
            return self.data[k][-1]
        return None

    def extend(self, other: 'MultiMap') -> Self:
        for k, v in other.pairs():
            self.add(k, v)
        return self

    def sortkeys(self, keys: Iterable[K] | None = None) -> Self:
        """Put the map keys in alphabetical order, or specified order."""
        if keys is None:
            keys = sorted(self.data.keys())
        old = self.data
        self.data = defaultdict(list[V])
        for k in keys:
            if k in old:
                self.data[k] = old[k]
                del old[k]
        self.data.update(old)
        return self

    def sortvalues(self,
                   keys: Iterable[K] | None = None,
                   key: Callable | None = None) -> Self:
        if keys is None:
            keys = self.data.keys()
        for k in keys:
            self.data[k].sort(key=key)  # type: ignore[reportGeneralTypeIssues]
        return self

    def submap(self, keys: Iterable[K] | None = None) -> Self:
        if not keys:
            return self
        r = type(self)()
        for k in keys or []:
            if k in self.data:
                r.data[k] = self.data[k]
        return r



# SPDX-License-Identifier: MIT
"""Timestamp."""

from fnattr.util.duration import Duration
from fnattr.vlju import Vlju

class Timestamp(Vlju):
    """
    Represents a timestamp.

    short:  [[[[[[_d_`:`]_h_]_h_`:`]_m_]_m_`:`]_s_]_s_[`.`_f_...]
    long:   [[[[[[_d_`d`]_h_]_h_`h`]_m_]_m_`m`]_s_]_s_[`.`_f_...]
    """

    def __init__(self, s: object) -> None:
        if isinstance(s, str):
            self._duration = Duration.parse(s)
        elif isinstance(s, Duration):
            self._duration = s
        elif hasattr(s, 'cast_params'):
            v, d = s.cast_params(type(self))
            self._duration = Duration.parse(v) if v else Duration(**d)
        else:
            raise TypeError(s)
        super().__init__(str(self._duration))

    def lv(self) -> str:
        return self._duration.fmt()

    def duration(self) -> Duration:
        return self._duration



# SPDX-License-Identifier: MIT
"""I/O utilities."""

import contextlib
import io
import os
import sys

from dataclasses import dataclass
from pathlib import Path
from typing import IO, cast

# Paths and open files are accepted.
PathLike = io.IOBase | IO | os.PathLike | str

@dataclass
class IOState:
    """Hold IO and whether we opened it."""

    file: IO
    opened: bool

def opener(file: PathLike | None,
           mode: str,
           default: IO,
           encoding: str = 'utf-8',
           **kwargs) -> IOState:
    """
    Open a file.

    The file can be an IOBase, '-' for the default, or a path.

    Returns a tuple of the file and a flag indicating whether the file
    was opened by this function.
    """
    if file is None or file == '-':
        return IOState(default, opened=False)
    if isinstance(file, io.IOBase | IO):
        return IOState(cast(IO, file), opened=False)
    return IOState(
        Path(file).open(mode, encoding=encoding, **kwargs), opened=True)

def open_context(file: PathLike | None,
                 mode: str,
                 default: IO,
                 encoding: str = 'utf-8',
                 **kwargs) -> contextlib.AbstractContextManager:
    s = opener(file, mode, default, encoding, **kwargs)
    if s.opened:
        return contextlib.closing(s.file)
    return contextlib.nullcontext(s.file)

def open_output(file: PathLike | None,
                default: IO = sys.stdout,
                encoding: str = 'utf-8',
                **kwargs) -> contextlib.AbstractContextManager:
    return open_context(file, 'w', default, encoding, **kwargs)

def open_input(file: PathLike | None,
               default: IO = sys.stdin,
               encoding: str = 'utf-8',
               **kwargs) -> contextlib.AbstractContextManager:
    return open_context(file, 'r', default, encoding, **kwargs)



# SPDX-License-Identifier: MIT
"""Configuration-file related utilities."""

import contextlib
import os
import tomllib

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Self

class Dirs(list[Path]):
    """Maintain a list of directories."""

    def add_env_dir(self, evar: str, default: Path | None = None) -> Self:
        """
        Add an environment-variable-based or default directory.

        Based on XDG conventions. If the environment variable named `evar`
        exists and contains a directory path, add it; otherwise, if the
        `default` exists, add that.
        """
        if evar in os.environ:
            p = Path(os.environ[evar])
            if p.is_dir():
                self.append(p)
        elif default and default.is_dir():
            self.append(default)
        return self

    def add_env_dirs(self, env: str, default: list[Path]) -> Self:
        """
        Add a list of environment-variable-based or default directories.

        Based on XDG conventions. If the environment variable named `evar`
        exists, treat it as a `:`-separated path and add any existing absolute
        directory. Otherwise, add any existing directory in `default`.
        """
        if env in os.environ:
            for d in os.environ[env].split(':'):
                p = Path(d)
                if p.is_dir() and p.is_absolute():
                    self.append(p)
        else:
            for p in default:
                if p.is_dir():
                    self.append(p)
        return self

    def find_first(self, p: Path | str) -> Path | None:
        """Return the first matching file in the directory list."""
        for i in self:
            d = i / p
            if d.exists():
                return d
        return None

def xdg_dirs(name: str,
             default_dir: str,
             default_paths: list[Path],
             home: Path | None = None) -> Dirs:
    """Obtain a list of XDG directories of the given kind."""
    if home is None:
        with contextlib.suppress(RuntimeError):
            home = Path.home()

    default_path = None if home is None else home / default_dir
    d = Dirs()
    d.add_env_dir(f'XDG_{name}_HOME', default_path)
    d.add_env_dirs(f'XDG_{name}_DIRS', default_paths)
    return d

def xdg_config_dirs() -> Dirs:
    return xdg_dirs('CONFIG', '.config', [Path('/etc/xdg')])

def xdg_config(filename: Path | str) -> Path | None:
    return xdg_config_dirs().find_first(filename)

def read_toml_config(file: Path | str) -> dict | None:
    with Path(file).open('rb') as f:
        try:
            return tomllib.load(f)
        except tomllib.TOMLDecodeError as e:
            print(f'{file}: {e}')
            return None

def read_configs(args: Iterable[Path | str]) -> dict:
    config: dict[str, Any] = {}
    for p in args:
        if c := read_toml_config(p):
            nested_update(config, c)
    return config

def read_xdg_configs(files: Iterable[Path]) -> dict:
    config: dict[str, Any] = {}
    for file in files:
        if (cf := xdg_config(file)) and (c := read_toml_config(cf)):
            nested_update(config, c)
    return config

def read_cmd_configs(cmd: str, args: Iterable[Path | str]) -> dict:
    if args:
        return read_configs(args)
    return read_xdg_configs([
        Path('vlju/config.toml'),
        Path(f'{cmd}/config.toml'),
    ])

def nested_update(dst: dict, src: dict) -> dict:
    for k, v in src.items():
        if k in dst and isinstance(dst[k], dict) and isinstance(v, dict):
            nested_update(dst[k], v)
        else:
            dst[k] = v
    return dst



# SPDX-License-Identifier: MIT
"""Vlju - top-level of the Vlju hierarchy."""

from fnattr.util.repr import mkrepr

class Vlju:
    """
    Vlju - Top level of the Vlju hierarchy.

    A Vlju is at minimum representable as a string.
    Subclasses may have additional structure.

    Subclasses constructors must take a single positional argument
    and accept at least a string. The canonical (short) representation
    produced by `__str__()` _must_ be accepted; the long representation
    produced by `lv()` _should_ be accepted.

    short:  value
    long:   value
    where:
        value → `_value`
    """

    def __init__(self, s: str) -> None:
        if not isinstance(s, str):
            raise TypeError(s)
        self._value = s

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Vlju):
            return self._value == other._value  # noqa: SLF001
        return False

    def __repr__(self) -> str:
        return mkrepr(self, ['_value'])

    def __str__(self) -> str:
        """Return the canonical (short) value."""
        return self._value

    def lv(self) -> str:
        """Return the long value."""
        return self._value

    def get(self,
            key: str | None = None,
            default: str | None = None) -> str | None:
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: str | None) -> str:
        match key:
            case None | 'default' | 'short' | 'str':
                return str(self)
            case 'long' | 'alternate':
                return self.lv()
            case 'repr':
                return repr(self)
        raise KeyError(key)

    def cast_param_error(self, t: object) -> TypeError:
        return TypeError((self, t))



# SPDX-License-Identifier: MIT
"""VljuMap operations."""

import copy
import re
import sys

from collections import defaultdict
from collections.abc import Mapping, MutableMapping
from pathlib import Path
from typing import Any, Self, TextIO

from fnattr.util.error import Error
from fnattr.util.io import PathLike, open_input, open_output
from fnattr.util.registry import Registry
from fnattr.vlju.types.all import URI, URL, File, Vlju
from fnattr.vljumap import VljuFactory, VljuMap, enc

VljuArg = Vlju | str | None
EncoderArg = enc.Encoder | str | None
FactoryArg = VljuFactory | str | None
ModeArg = str | None

class VljuM(VljuMap):
    """VljuMap operations."""

    default_registry: MutableMapping[str, Registry] = defaultdict(Registry)

    def __init__(self, i: VljuMap | File | Path | str | object = None) -> None:
        super().__init__()
        self.factory = copy.copy(self.default_registry['factory'])
        self.encoder = copy.copy(self.default_registry['encoder'])
        self.decoder = copy.copy(self.default_registry['decoder'])
        self.mode = copy.copy(self.default_registry['mode'])
        self._original_path: Path
        self._current_dir: Path
        self._current_suffix: str
        self._set_path(Path())
        if i is not None:
            if isinstance(i, VljuMap):
                self.extend(i)
            elif isinstance(i, File):
                self.file(i.filename())
            elif isinstance(i, Path):
                self._set_path(i)
            elif isinstance(i, str):
                self.decode(i)
            elif hasattr(i, 'cast_params'):
                s, d = i.cast_params(type(self))
                if s:
                    self._set_path(s)
                for k, v in d.items():
                    self.add(k, v)
            else:
                raise TypeError(i)

    def cast_params(self, t: object) -> tuple[str | Path, dict]:
        if t is File:
            return (self.filename(), {})
        if t is URI or t is URL:
            return (str(self.filename()), {'scheme': 'file', 'sa': '://'})
        raise TypeError((self, t))

    # Map operations.

    def add(self,
            k: str,
            v: VljuArg = None,
            factory: FactoryArg = None) -> Self:
        super().add(k, self._vlju(k, v, factory))
        return self

    def decode(self,
               s: str,
               decoder: EncoderArg = None,
               factory: FactoryArg = None) -> Self:
        self.decoder.get(decoder).decode(self, s, self.factory.get(factory))
        return self

    def extract(self, *args: str) -> Self:
        return self.submap(args)

    def file(self,
             s: str | Path,
             decoder: EncoderArg = None,
             factory: FactoryArg = None) -> Self:
        self._set_path(s)
        self.decoder.get(decoder).decode(self,
                                         self._original_path.stem,
                                         self.factory.get(factory))
        return self

    def order(self, *args: str) -> Self:
        return self.sortkeys(args or None)

    def read(self,
             file: PathLike = '-',
             decoder: EncoderArg = None,
             factory: FactoryArg = None) -> Self:
        with open_input(file, sys.stdin) as f:
            self.decoder.get(decoder).decode(self,
                                             f.read(),
                                             self.factory.get(factory))
        return self

    def remove(self,
               k: str,
               v: VljuArg = None,
               factory: FactoryArg = None) -> Self:
        if v is None:
            del self[k]
            return self
        super().remove(k, self._vlju(k, v, factory))
        return self

    def rename(self, encoder: EncoderArg = None, mode: ModeArg = None) -> Self:
        if self._original_path == Path():
            message = 'no file to rename'
            raise Error(message)
        modified_path = self.filename(encoder, self.mode.get(mode))
        if modified_path.exists():
            if modified_path.samefile(self._original_path):
                return self
            raise FileExistsError(modified_path)
        self._original_path.rename(modified_path)
        self._set_path(modified_path)
        return self

    def reset(self,
              k: str,
              v: VljuArg = None,
              factory: FactoryArg = None) -> Self:
        del self[k]
        return self.add(k, v, factory)

    def sort(self, *args: str, mode: ModeArg = None) -> Self:
        return self.sortvalues(args or None,
                               lambda v: v.get(self.mode.get(mode)))

    def with_dir(self, s: str | Path) -> Self:
        self._current_dir = Path(s)
        return self

    def with_suffix(self, suffix: str) -> Self:
        if not suffix.startswith('.'):
            suffix = '.' + suffix
        self._current_suffix = suffix
        return self

    def write(self,
              file: PathLike = '-',
              encoder: EncoderArg = None,
              mode: ModeArg = None) -> Self:
        with open_output(file, sys.stdout) as f:
            f.write(self.encoder.get(encoder).encode(self, self.mode.get(mode)))
        return self

    def z(self, file: TextIO = sys.stderr) -> Self:     # pragma: no coverage
        print(repr(self), file=file)
        return self

    # String reductions.

    def __str__(self) -> str:
        return str(self.filename())

    def lv(self) -> str:
        return self.encode(mode='long')

    def encode(self, encoder: EncoderArg = None, mode: ModeArg = None) -> str:
        return self.encoder.get(encoder).encode(self, self.mode.get(mode))

    def collect(self,
                *args: str,
                encoder: EncoderArg = None,
                mode: ModeArg = None) -> str:
        return self.encoder.get(encoder).encode(
            self.submap(args), self.mode.get(mode))

    def q(self) -> str:
        return ''

    def uri(self, encoder: EncoderArg = 'value') -> str:
        return self._url(URI).encode(encoder)

    def url(self, encoder: EncoderArg = 'value') -> str:
        return self._url(URL).encode(encoder)

    # Filename reduction.

    def original(self) -> Path:
        return self._original_path

    def filename(self,
                 encoder: EncoderArg = None,
                 mode: ModeArg = None) -> Path:
        e = self.encode(encoder, self.mode.get(mode))
        if not e:
            message = 'no file name'
            raise Error(message)
        return (self._current_dir / e).with_suffix(self._current_suffix)

    # Vlju reduction.

    def first(self, k: str | type[Vlju]) -> Vlju:
        if isinstance(k, str):
            if k in self:
                return self[k][0]
        else:
            for _, v in self.pairs():
                if isinstance(v, k):
                    return v
        return Vlju('')

    # Helpers.

    def _set_path(self, s: str | Path) -> Self:
        if isinstance(s, str):
            s = Path(s)
        self._original_path = s
        self._current_dir = s.parent
        self._current_suffix = s.suffix
        return self

    def _url(self, cls: type) -> Self:
        # Try hard to get URIs/URLs from the current map.
        out = type(self)()
        strings: list[tuple[str, str]] = []
        for k, v in self.pairs():
            try:
                u = cls(v)
            except TypeError:
                u = None
            if u:
                out.add(k, u)
            else:
                strings.append((k, str(v)))
        if strings:
            scheme_slashes_re = re.compile(r'\w+://.+')
            for k, v in strings:
                if '/' not in v or scheme_slashes_re.fullmatch(v):
                    t = v
                else:
                    t = 'http://' + v
                u = cls(t)
                if u and hasattr(u, 'authority') and u.authority():
                    out.add(k, u)
        return out

    def _vlju(self, k: str, v: VljuArg, factory: FactoryArg = None) -> Vlju:
        if v is None:
            r = Vlju('')
        elif isinstance(v, str):
            _, r = self.factory.get(factory)(k, v)
        else:
            r = v
        return r

    def __repr__(self) -> str:
        return f'{type(self).__name__}({dict(self.data)!r})'

    @classmethod
    def configure_options(cls, options: Mapping[str, Any]) -> None:
        for r in cls.default_registry:
            if (v := options.get(r)) is not None:
                cls.default_registry[r].set_default(v)



# SPDX-License-Identifier: MIT
"""Info - an entity in the ‘info’ URI scheme."""

from fnattr.vlju.types.uri import URI, auth

class Info(URI):
    """
    Represents an entity in the ‘info’ URI scheme.

    short:  uri
    long:   uri
    """

    def __init__(self, s: str, **kwargs) -> None:
        super().__init__(
            s,
            scheme='info',
            authority=auth(kwargs.get('authority')),
            sa=':',
            ap='/')

    def __str__(self) -> str:
        return f'{self.sauthority()}/{self._value}'

    def sauthority(self) -> str:
        a = self.authority()
        return str(a.host) if a else ''

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URI:
            return (self._value,
                    {
                        'scheme': self._scheme,
                        'authority': self._authority,
                        'sa': ':',
                    })
        raise self.cast_param_error(t)



# SPDX-License-Identifier: MIT
"""Wrapper including context manager for sqlite3."""

import os
import sqlite3

from collections.abc import Iterable
from types import TracebackType
from typing import Any, Self

Connection = sqlite3.Connection
Cursor = sqlite3.Cursor
PathLike = os.PathLike | str

SQLITE3_OPEN_QUERY_KEYS = {
    'cache',
    'immutable',
    'mode',
    'modeof',
    'nolock',
    'psow',
    'vgs',
}

class SQLite:
    """Wrapper including context manager for sqlite3."""

    # These are intended for subclasses to provide initialization for all
    # instances of the subclasses. The model is that a subclass defines a
    # database including a schema definition as its `on_create`.
    on_connect: Iterable[str] | None = None
    on_create: Iterable[str] | None = None

    def __init__(self,
                 filename: PathLike = '',
                 mode: str = '',
                 **kwargs) -> None:
        if not mode:
            mode = 'ro' if filename else 'rw'
        self._filename: PathLike = filename
        self._kwargs = kwargs
        self._kwargs['mode'] = mode
        self._connection: Connection | None = None
        self._table_columns: dict[str, list[str] | None] = {}

    def __enter__(self) -> Self:
        return self.connect()

    def connect(self) -> Self:
        """Open and initialize the database connection."""
        if not self._connection:
            self._connection, created = self._open()
            if self.on_connect:
                for i in self.on_connect:
                    self._connection.execute(i)
            if created and self.on_create:
                for i in self.on_create:
                    self._connection.execute(i)
        return self

    def _open(self) -> tuple[Connection, bool]:
        if not self._filename or self._kwargs['mode'] == 'memory':
            # Temporary files are always created.
            created = True
        elif self._kwargs['mode'] == 'rwc':
            # Try first with 'rw', to see whether it needs to be created.
            self._kwargs['mode'] = 'rw'
            try:
                con = sqlite3.connect(self._uri(), uri=True)
            except sqlite3.OperationalError:
                pass
            else:
                return con, False
            self._kwargs['mode'] = 'rwc'
            created = True
        else:
            created = False
        return sqlite3.connect(self._uri(), uri=True), created

    def _uri(self) -> str:
        q = '&'.join(f'{k}={self._kwargs[k]!s}' for k in self._kwargs.keys()
                     & SQLITE3_OPEN_QUERY_KEYS)
        return f'file:{self._filename!s}?{q}'

    def __exit__(self,
                 et: type[BaseException],
                 ev: BaseException,
                 traceback: TracebackType) -> None:
        self.close()

    def close(self) -> Self:
        if self._connection is not None:
            self._connection.close()
            self._connection = None
        self.clear_table_column_cache()
        return self

    def connection(self) -> Connection:
        if not self._connection:
            raise sqlite3.OperationalError
        return self._connection

    def commit(self) -> Self:
        self.connection().commit()
        return self

    def execute(self, query: str, *args, **kwargs) -> Cursor:
        if args and kwargs:
            message = 'cannot use both positional and keyword arguments'
            raise ValueError(message)
        if args:
            return self.connection().execute(query, args)
        if kwargs:
            return self.connection().execute(query, kwargs)
        return self.connection().execute(query)

    def store(self,
              table: str,
              on_conflict: str | None = None,
              **kwargs) -> Self:
        """
        Insert into a table.

        `kwargs` consists of column-value pairs.

        This is intended as a shorthand to be used with hard-coded keys,
        for example:

            db.store(users', user=u, password=p)

        However, due to splatting, it is possible that this function could
        be (mis)used with untrusted arguments. Therefore, we take precautions
        against injection, since table and column names must be part of the
        query string:

        - Verify that the supplied table and column names are actually
          existing table and column names in the database.
        - Quote table and column names.
        - Use generated parameter names.
        """
        columns = kwargs.keys()
        self.check_table_columns(table, kwargs.keys())
        p = BoundParameters(kwargs)
        q = (
            f'INSERT INTO {quote_id(table)}'                # noqa: S608
            f' ({",".join(quote_id(c) for c in columns)})'
            f' VALUES ({",".join(f":{k}" for k in p.value)})')
        if on_conflict:
            q += ' ON CONFLICT ' + on_conflict
        self.connection().execute(q, p.value)
        return self

    def load(self, table: str, *args: str, **kwargs) -> Cursor:
        """
        Read from a table.

        `args` consists of column names to be returned.

        `kwargs` consists of column-value pairs that become `WHERE`
        conditions on the select.

        Although this is not intended to handle untrusted arguments,
        we take precautions against injection, since table and column
        names must be part of the query string:

        - Verify that the supplied table and column names are actually
          existing table and column names in the database.
        - Quote table and column names.
        - Use generated parameter names.
        """
        check_columns = set(kwargs.keys())
        if args:
            check_columns |= set(args)
            cols = ','.join(quote_id(c) for c in args)
        else:
            cols = '*'
        self.check_table_columns(table, check_columns)
        q = f'SELECT {cols} FROM {quote_id(table)}'     # noqa: S608
        if not kwargs:
            return self.connection().execute(q)
        p = BoundParameters(kwargs)
        q += ' WHERE ' + ' AND '.join(f'{p.column[k]} = :{k}' for k in p.value)
        return self.connection().execute(q, p.value)

    def check_table_columns(self, table: str, columns: Iterable[str]) -> None:
        """Check that the columns exist in the table."""
        if not self._table_columns:
            self._init_table_column_cache()
        if table not in self._table_columns:
            message = f'table {table!r} does not exist'
            raise sqlite3.ProgrammingError(message)
        if (table_columns := self._table_columns[table]) is None:
            table_columns = self._load_table_column_cache(table)
        for c in columns:
            if c not in table_columns:
                message = f'column {c!r} does not exist in {table!r}'
                raise sqlite3.ProgrammingError(message)

    def _init_table_column_cache(self) -> None:
        c = self.connection().execute(
            'SELECT name FROM sqlite_master WHERE type == "table"')
        while (row := c.fetchone()):
            table = row[0]
            self._table_columns[table] = None

    def _load_table_column_cache(self, table: str) -> list[str]:
        r = []
        c = self.connection().execute(f'PRAGMA TABLE_INFO({quote_id(table)})')
        while (row := c.fetchone()):
            column = row[1]
            r.append(column)
        self._table_columns[table] = r
        return r

    def clear_table_column_cache(self) -> None:
        self._table_columns = {}

def quote_id(s: str) -> str:
    return '"' + s.replace('"', '""') + '"'

class BoundParameters:
    """Sanitization of bound parameters."""

    column: dict[str, str]  # Map from parameter name to quoted column name.
    value: dict[str, Any]   # Map from parameter name to value.

    def __init__(self, column_value: dict[str, Any]) -> None:
        self.column = {}
        self.value = {}
        for i, k in enumerate(column_value):
            p = f'p{i}'
            self.column[p] = quote_id(k)
            self.value[p] = column_value[k]



# SPDX-License-Identifier: MIT
"""`fna` command."""

import argparse
import pathlib
import sys

import fnattr.util.config
import fnattr.util.io
import fnattr.vljum.m
import fnattr.vljum.runner
import fnattr.vljumap.enc

def main(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv
    cmd = pathlib.Path(argv[0]).stem
    parser = argparse.ArgumentParser(
        prog=cmd,
        description='Manage key/value attributes in file names',
        epilog=(f'For a list of subcommands, run `{cmd} help`. '
                f'For information on a specific subcommand, '
                f'run `{cmd} help ‹subcommand›`'))
    parser.add_argument(
        '--config',
        '-c',
        metavar='FILE',
        type=str,
        action='append',
        help='Configuration file.')
    parser.add_argument(
        '--decoder',
        '-d',
        metavar='DECODER',
        type=str,
        choices=fnattr.vljumap.enc.decoder.keys(),
        help='Default string decoder.')
    parser.add_argument(
        '--encoder',
        '-e',
        metavar='ENCODER',
        type=str,
        choices=fnattr.vljumap.enc.encoder.keys(),
        help='Default string encoder.')
    mode = parser.add_mutually_exclusive_group()
    mode.add_argument(
        '--dsl',
        '-D',
        dest='mode',
        default='dsl',
        const='dsl',
        action='store_const',
        help='Positional arguments are subcommands (default).')
    mode.add_argument(
        '--evaluate',
        '-E',
        dest='mode',
        const='evaluate',
        action='store_const',
        help='Positional arguments are expressions to evaluate.')
    mode.add_argument(
        '--execute',
        '-x',
        dest='mode',
        const='execute',
        action='store_const',
        help='Positional arguments are statements to execute.')
    mode.add_argument(
        '--file',
        '-f',
        dest='mode',
        const='file',
        action='store_const',
        help='Positional arguments are program files.')
    parser.add_argument(
        'argument',
        metavar='ARGUMENT',
        type=str,
        nargs=argparse.REMAINDER,
        default=[],
        help='Part of a subcommand, or an expression or statement.')
    args = parser.parse_args(argv[1 :])

    config = fnattr.util.config.read_cmd_configs(cmd, args.config)
    options = config.get('option', {})
    if args.decoder:
        options['decoder'] = args.decoder
    if args.encoder:
        options['encoder'] = args.encoder
    fnattr.vljum.m.M.configure_options(options)
    fnattr.vljum.m.M.configure_sites(config.get('site', {}))

    try:
        match args.mode:
            case 'dsl':
                fnattr.vljum.runner.Runner().run(args.argument)
            case 'evaluate':
                for i in args.argument:
                    r = fnattr.vljum.m.M.evaluate(i)
                    if r is not None:
                        print(r)
            case 'execute':
                for i in args.argument:
                    fnattr.vljum.m.M.execute(i)
            case 'file':
                for i in args.argument:
                    with fnattr.util.io.open_input(i) as f:
                        fnattr.vljum.m.M.execute(f.read())
            case _:
                print(f'{cmd}: Unknown mode: {args.mode}')
    except Exception as e:
        print(f'{cmd}: Unhandled exception: {type(e).__name__}{e.args}')
        raise
    return 0

if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: MIT
"""URI - Vlju representable as a URI."""

import re

from typing import Any, Self

from fnattr.util import escape
from fnattr.util.repr import mkrepr
from fnattr.util.typecheck import needtype
from fnattr.vlju import Vlju

class Authority:
    """Authority represents a URI authority."""

    host: str
    port: int | None = None
    username: str | None = None
    password: str | None = None

    def __init__(self,
                 host: str | Self,
                 port: int | None = None,
                 username: str | None = None,
                 password: str | None = None) -> None:

        if isinstance(host, Authority):
            self.username = host.username
            self.password = host.password
            self.host = host.host
            self.port = host.port
        elif isinstance(host, str):
            if host.startswith('//'):
                host = host[2 :]
            if '@' in host:
                up, host = host.split('@', 1)
                if ':' in up:
                    u, p = up.split(':', 1)
                else:
                    u = up
                    p = None
                if username is not None and username != u:
                    message = f'username={username} conflicts with {up}@'
                    raise ValueError(message)
                if password is not None and password != p:
                    message = f'password={password} conflicts with {up}@'
                    raise ValueError(message)
                self.username = u
                self.password = p
            if ':' in host:
                host, pstr = host.split(':', 1)
                pn = int(pstr)
                if port is not None and port != pn:
                    message = f'port={port} conflicts with {host}'
                    raise ValueError(message)
                self.port = pn
            self.host = host.lower()
        else:
            raise TypeError

        if port is not None:
            self.port = port
        if username is not None:
            self.username = username
        if password is not None:
            self.password = password

    def __repr__(self) -> str:
        return mkrepr(self, ['host'], ['port', 'username', 'password'])

    def __str__(self) -> str:
        r = ''
        if self.username:
            r += escape.auth.encode(self.username)
            if self.password:
                r += ':' + escape.auth.encode(self.password)
            r += '@'
        r += escape.auth.encode(self.host)
        if self.port:
            r += f':{self.port}'
        return r

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Authority):
            return (self.host == other.host and self.port == other.port
                    and self.username == other.username
                    and self.password == other.password)
        return False

AuthorityArg = Authority | str | None

def auth(a: AuthorityArg) -> Authority | None:
    if a is None:
        return None
    if isinstance(a, Authority):
        return a
    return Authority(a)

class URI(Vlju):
    """
    Represents a URI.

    short:  uri
    long:   uri
    where:
        uri   → `scheme()` `_sa` `sauthority()` `_ap` `spath()`
                `squery()` `sfragment()` `sr()` `sq()`
    """

    def __init__(self, s: str | object, **kwargs) -> None:
        if isinstance(s, str):
            if kwargs:
                v: str = s
            else:
                v, kwargs = split_uri(s)
        elif hasattr(s, 'cast_params'):
            v, d = s.cast_params(type(self))
            kwargs = d | kwargs
        else:
            raise TypeError(s)
        super().__init__(v)
        self._scheme: str = needtype(kwargs.get('scheme', ''), str)
        self._authority: Authority | None = auth(kwargs.get('authority'))
        self._query: str | None = needtype(kwargs.get('query'), str, None)
        self._fragment: str | None = needtype(kwargs.get('fragment'), str, None)
        self._urnq: str | None = needtype(kwargs.get('urnq'), str, None)
        self._urnr: str | None = needtype(kwargs.get('urnr'), str, None)
        sa = kwargs.get('sa')   # scheme/authority separator
        if sa is None:
            sa = ':' if self._scheme else ''
            sa += '//' if self._authority else ''
        self._sa: str = needtype(sa, str)
        ap = kwargs.get('ap')   # authority/path separator
        if ap is None:
            ap = '/' if (self._authority and self._value[0].isalnum()) else ''
        self._ap: str = needtype(ap, str)

    def scheme(self) -> str:
        return self._scheme

    def path(self) -> str:
        return self._value

    def authority(self) -> Authority | None:
        return self._authority

    def query(self) -> str | None:
        return self._query

    def q(self) -> str | None:
        return self._urnq

    def r(self) -> str | None:
        return self._urnr

    def fragment(self) -> str | None:
        return self._fragment

    def spath(self) -> str:
        return f'{escape.path.encode(self.path())}'

    def sauthority(self) -> str:
        a = self.authority()
        return str(a) if a else ''

    def squery(self) -> str:
        s = self.query()
        return f'?{escape.query.encode(s)}' if s else ''

    def sq(self) -> str:
        s = self.q()
        return f'?={escape.query.encode(s)}' if s else ''

    def sr(self) -> str:
        s = self.r()
        return f'?+{escape.query.encode(s)}' if s else ''

    def sfragment(self) -> str:
        s = self.fragment()
        return f'#{escape.fragment.encode(s)}' if s else ''

    def uri(self, path: str | None = None) -> str:
        return (self.scheme() + self._sa + self.sauthority() + self._ap +
                (self.spath() if path is None else escape.path.encode(path))
                + self.squery() + self.sfragment() + self.sr() + self.sq())

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URI:
            return (self.path(),
                    {
                        'scheme': self._scheme,
                        'authority': self._authority,
                        'query': self._query,
                        'fragment': self._fragment,
                        'urnq': self._urnq,
                        'urnr': self._urnr,
                        'sa': self._sa,
                        'ap': self._ap,
                    })
        raise self.cast_param_error(t)

    # Vlju overrides:

    def __str__(self) -> str:
        return self.uri()

    def lv(self) -> str:
        return self.uri()

    def __getitem__(self, key: str | None) -> str:
        if key in ('scheme',
                   'path',
                   'authority',
                   'query',
                   'fragment',
                   'spath',
                   'sscheme',
                   'sauthority',
                   'squery',
                   'sfragment'):
            return getattr(self, key)()
        return super().__getitem__(key)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, URI):
            return (self._value == other._value                 # noqa: SLF001
                    and self._scheme == other._scheme           # noqa: SLF001
                    and self._authority == other._authority     # noqa: SLF001
                    and self._query == other._query             # noqa: SLF001
                    and self._fragment == other._fragment       # noqa: SLF001
                    and self._urnq == other._urnq               # noqa: SLF001
                    and self._urnr == other._urnr               # noqa: SLF001
                    and self._sa == other._sa                   # noqa: SLF001
                    and self._ap == other._ap)                  # noqa: SLF001
        return False

    def __repr__(self) -> str:
        return mkrepr(          # pragma: no cover
            self, ['_value'],
            ['_scheme', '_sa', '_authority', '_query', '_fragment'])

SCHEME_RE = re.compile(r'(?P<scheme>\w+):(?P<rest>.+)')

def split_uri(s: str) -> tuple[str, dict[str, Any]]:
    """Split a URI string."""
    d = {}

    # scheme
    if m := SCHEME_RE.fullmatch(s):
        d['scheme'], s = m.group('scheme', 'rest')

    # authority
    if s.startswith('//') and (i := s.find('/', 2)) > 0:
        d['authority'] = Authority(s[2 : i])
        s = s[i :]

    # fragment
    if (i := s.find('#')) > 0:
        d['fragment'] = s[i + 1 :]
        s = s[: i]

    # query
    if (i := s.find('?')) > 0:
        d['query'] = s[i + 1 :]
        s = s[: i]

    return (s, d)



# SPDX-License-Identifier: MIT
"""Checksums."""

def alt13_checksum(s: str) -> int:
    """Calculate an EAN checksum."""
    r = -10
    multiplier = 1
    for c in s:
        r += multiplier * int(c)
        multiplier = 4 - multiplier
    return (10 - r % 10) % 10

def alt13_checksum_to_check_digit(n: int) -> str:
    return str(n)

def alt13(s: str) -> str:
    return alt13_checksum_to_check_digit(alt13_checksum(s))

def mod11_checksum(s: str) -> int:
    """Calculate an ISBN-10 checksum."""
    r = 0
    for n in range(1, 1 + len(s)):
        r += (n + 1) * int(s[-n])
    r = 11 - r % 11
    return r

def mod11_checksum_to_check_digit(n: int) -> str:
    return '0123456789X'[n]

def mod11(s: str) -> str:
    return mod11_checksum_to_check_digit(mod11_checksum(s))



# SPDX-License-Identifier: MIT
"""VljuMap."""

from collections.abc import Generator, Iterable
from typing import Self

from fnattr.util.multimap import MultiMap
from fnattr.vlju import Vlju
from fnattr.vljumap.factory import VljuFactory, default_factory

class VljuMap(MultiMap[str, Vlju]):
    """MultiMap of Vlju."""

    def get_pairs(
        self,
        mode: str | None = None,
    ) -> Generator[tuple[str, str | None], None, None]:
        for k, v in self.pairs():
            yield (k, v.get(mode))

    def get_lists(
        self,
        mode: str | None = None,
    ) -> Generator[tuple[str, list[str | None]], None, None]:
        for k, vlist in self.lists():
            yield (k, [v.get(mode) for v in vlist])

    def to_strings(self, mode: str | None = None) -> MultiMap[str, str]:
        m: MultiMap[str, str] = MultiMap()
        for k, v in self.get_pairs(mode):
            m.add(k, str(v))
        return m

    def add_pairs(self,
                  i: Iterable[tuple[str, str]],
                  factory: VljuFactory | None = None) -> Self:
        if factory is None:
            factory = default_factory
        for k, v in i:
            self.add(*factory(k, v))
        return self



# SPDX-License-Identifier: MIT
"""Split a docstring."""

import textwrap

def docsplit(s: str) -> tuple[list[str], dict[str, str]]:
    """
    Split off keyword sections.

    Puts paragraphs that begin with a word followed by `:` into the dictionary
    result, and others into the list result.
    """
    r = []
    d = {}
    for i in s.split('\n\n'):
        t = textwrap.dedent(i).strip()
        if ((n := t.find(':')) > 0) and t[: n].isalpha():
            k, v = t.split(':', 1)
            d[k.lower()] = v.strip()
            continue
        t = t.strip()
        if t:
            r.append(t)
    return (r, d)



# SPDX-License-Identifier: MIT
"""Like format, but scary."""

import operator

from collections.abc import Mapping
from typing import Any

from fnattr.util.error import Error

ALLOWED_BUILTINS = (
    'False',
    'None',
    'True',
    'abs',
    'all',
    'any',
    'ascii',
    'bin',
    'bool',
    'chr',
    'hex',
    'int',
    'len',
    'map',
    'max',
    'min',
    'oct',
    'ord',
    'reversed',
    'slice',
    'sorted',
    'str',
)
ALLOWED_OPERATORS = (
    'add',
    'sub',
)

BUILTINS = {
    k: globals()['__builtins__'][k]
    for k in ALLOWED_BUILTINS
} | {
    k: getattr(operator, k)
    for k in ALLOWED_OPERATORS
}

def fearmat(template: str,
            values: Mapping[str, Any],
            builtins: Mapping[str, Any] | None = None) -> str:
    if '"""' in template:
        msg = '‘"""’ in ‘template’'
        raise Error(msg)
    template = 'f"""' + template + '"""'
    if builtins is None:
        builtins = BUILTINS
    g = dict(values) | {'__builtins__': builtins}
    return str(eval(template, g))   # noqa: PGH001, eval-used



# SPDX-License-Identifier: MIT
"""Runtime type checking utilities."""

from typing import TypeVar

T = TypeVar('T')

def istype(value, *args) -> bool:   # noqa: ANN001
    for i in args:
        if i is None:
            return value is None
        if isinstance(value, i):
            return True
    return False

def needtype(value: T, *args) -> T:
    if istype(value, *args):
        return value
    raise TypeError((value, args))



# SPDX-License-Identifier: MIT
"""Kinds of URN. Not implemented."""

from fnattr.vlju import Vlju

KIND: dict[str, type[Vlju] | None] = {}

ORGANIZATION: dict[str, str] = {}

# https://www.iana.org/assignments/urn-namespaces/urn-namespaces.xhtml

# 3gpp                https://www.iana.org/go/rfc5279
# 3gpp2               https://www.iana.org/go/rfc8464
# adid                https://www.iana.org/go/rfc8107
# alert               https://www.iana.org/go/rfc7462
# bbf                 https://www.iana.org/go/rfc8057
# broadband-forum-org https://www.iana.org/go/rfc8057
# cablelabs           https://www.iana.org/go/rfc6289
# ccsds               https://www.iana.org/go/rfc7738
# cgi                 https://www.iana.org/go/rfc5138
# clei                https://www.iana.org/go/rfc4152
# ddi                 https://www.iana.org/go/draft-urn-ddi-00
# dev                 https://www.iana.org/go/rfc9039
# dgiwg               https://www.iana.org/go/rfc6288
# dslforum-org        https://www.iana.org/go/rfc8057
# dvb                 https://www.iana.org/go/rfc7354
# ebu                 https://www.iana.org/go/rfc5174
# eidr                https://www.iana.org/go/rfc7972
# epc                 https://www.iana.org/go/rfc5134
# epcglobal           https://www.iana.org/go/rfc5134
# etsi                https://www.iana.org/go/rfc8515
# eurosystem          https://www.iana.org/go/rfc7207
# example             https://www.iana.org/go/rfc6963
# fdc                 https://www.iana.org/go/rfc4198
# fipa                https://www.iana.org/go/rfc3616
# geant               https://www.iana.org/go/rfc4926
# globus              https://www.iana.org/go/rfc7853
# gsma                https://www.iana.org/go/rfc7254
# hbbtv               https://www.iana.org/go/rfc7528
# ieee                https://www.iana.org/go/rfc8069
# ietf                https://www.iana.org/go/rfc2648
# iptc                https://www.iana.org/go/rfc3937
# isan                https://www.iana.org/go/rfc4246
# isbn                https://www.isbn-international.org
# iso                 https://www.iana.org/go/rfc5141
# issn                http://www.issn.org
# itu                 https://www.itu.int/
# ivis                https://www.iana.org/go/rfc4617
# liberty             https://www.iana.org/go/rfc3622
# mace                https://www.iana.org/go/rfc3613
# mef                 https://www.iana.org/go/rfc7818
# mpeg                https://www.iana.org/go/rfc3614
# mrn                 http://www.iala-aism.org
# nato                https://www.iana.org/go/rfc7467
# nbn                 https://www.iana.org/go/rfc8458
# nena                https://www.iana.org/go/rfc6061
# newsml              https://www.iana.org/go/rfc3085
# nfc                 https://www.iana.org/go/rfc4729
# nzl                 https://www.iana.org/go/rfc4350
# oasis               https://www.iana.org/go/rfc3121
# ogc                 https://www.iana.org/go/rfc5165
# ogf                 https://www.iana.org/go/rfc6453
# oid                 https://www.iana.org/go/rfc3061
# oipf                https://www.iana.org/go/rfc6893
# oma                 https://www.iana.org/go/rfc4358
# onf                 https://www.opennetworking.org/
# pin                 https://www.iana.org/go/rfc3043
# publicid            https://www.iana.org/go/rfc3151
# reso                http://www.iana.org/assignments/urn-formal/reso
# s1000d              https://www.iana.org/go/rfc4688
# schac               https://www.iana.org/go/rfc6338
# service             https://www.iana.org/go/rfc5031
# smpte               https://www.iana.org/go/rfc5119
# swift               https://www.iana.org/go/rfc3615
# tva                 https://www.iana.org/go/rfc4195
# uci                 https://www.iana.org/go/rfc4179
# ucode               http://www.rfc-editor.org/errata_search.php?eid=3189
# uuid                https://www.iana.org/go/rfc4122
# web3d               https://www.iana.org/go/rfc3541
# xmlorg              https://www.iana.org/go/rfc3120
# xmpp                https://www.iana.org/go/rfc4854



# SPDX-License-Identifier: MIT
"""Duration."""

# Why not just timedelta? I wanted nanosecond resolution like `struct timespec`.

import re

from datetime import timedelta
from typing import Self

NANOSECONDS_PER_DAY = 24 * 60 * 60 * 1_000_000_000
NANOSECONDS_PER_HOUR = 60 * 60 * 1_000_000_000
NANOSECONDS_PER_MINUTE = 60 * 1_000_000_000
NANOSECONDS_PER_SECOND = 1_000_000_000
NANOSECONDS_PER_MILLISECOND = 1_000_000
NANOSECONDS_PER_MICROSECOND = 1_000
NANOSECONDS_PER_NANOSECOND = 1

class Duration:
    """Duration."""

    def __init__(self,
                 *,
                 days: int | float = 0,
                 hours: int | float = 0,
                 minutes: int | float = 0,
                 seconds: int | float = 0,
                 milliseconds: int | float = 0,
                 microseconds: int | float = 0,
                 nanoseconds: int = 0) -> None:
        # Python has arbitrary-precision integers but not abritrary-precision
        # floats, so we convert each term to integer nanoseconds separately.
        self.ns: int = (
            nanoseconds + int(microseconds * NANOSECONDS_PER_MICROSECOND)
            + int(milliseconds * NANOSECONDS_PER_MILLISECOND)
            + int(seconds * NANOSECONDS_PER_SECOND)
            + int(minutes * NANOSECONDS_PER_MINUTE)
            + int(hours * NANOSECONDS_PER_HOUR)
            + int(days * NANOSECONDS_PER_DAY))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Duration):
            return self.ns == other.ns
        if isinstance(other, int):
            return self.ns == other
        return False

    def to_nanoseconds(self) -> int:
        return self.ns

    def to_seconds(self) -> float:
        return self.ns / NANOSECONDS_PER_SECOND

    def to_dhmsn(self) -> tuple[int, int, int, int, int]:
        return nanoseconds_to_dhmsn(self.ns)

    def to_timedelta(self) -> timedelta:
        return timedelta(microseconds=self.ns / NANOSECONDS_PER_MICROSECOND)

    def __repr__(self) -> str:
        return f'{type(self).__name__}(nanoseconds={self.ns})'

    def __str__(self) -> str:
        return self.fmt(':::.')

    def fmt(self, p: str = 'dhms') -> str:
        """Minimally format a duration."""
        return fmt_nanoseconds(self.ns, p)

    @classmethod
    def parse(cls, s: str) -> Self:
        return cls(nanoseconds=parse_to_nanoseconds(s))

    @classmethod
    def from_timedelta(cls, t: timedelta) -> Self:
        return cls(days=t.days, seconds=t.seconds, microseconds=t.microseconds)

def nanoseconds_to_dhmsn(ns: int) -> tuple[int, int, int, int, int]:
    s, n = divmod(ns, NANOSECONDS_PER_SECOND)
    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    return d, h, m, s, n

def fmt_nanoseconds(nanoseconds: int, p: str = 'dhms') -> str:
    """Minimally format a duration."""
    if nanoseconds < 0:
        sign = '-'
        nanoseconds = -nanoseconds
    else:
        sign = ''
    d, h, m, s, n = nanoseconds_to_dhmsn(nanoseconds)
    r = (
        f'{d}{p[0]}{h:02}{p[1]}{m:02}{p[2]}{s:02}'  #
        .removeprefix(f'0{p[0]}')                   # if no days
        .removeprefix('0')                          # if hours < 10
        .removeprefix(f'0{p[1]}')                   # if no hours
        .removeprefix('0')                          # if minutes < 10
        .removeprefix(f'0{p[2]}')                   # if no minutes
        .removeprefix('0')                          # if seconds < 10
    )
    ns = ''
    if n:
        ns = f'{n:09}'
        while (t := ns.removesuffix('000')) != ns:
            ns = t
    return f'{sign}{r}{p[3]}{ns}'.removesuffix('.')

__UNIT_TO_NANOSECONDS = {
    'd': NANOSECONDS_PER_DAY,
    'day': NANOSECONDS_PER_DAY,
    'days': NANOSECONDS_PER_DAY,
    'h': NANOSECONDS_PER_HOUR,
    'hour': NANOSECONDS_PER_HOUR,
    'hours': NANOSECONDS_PER_HOUR,
    'm': NANOSECONDS_PER_MINUTE,
    'minute': NANOSECONDS_PER_MINUTE,
    'minutes': NANOSECONDS_PER_MINUTE,
    '′': NANOSECONDS_PER_MINUTE,
    "'": NANOSECONDS_PER_MINUTE,
    's': NANOSECONDS_PER_SECOND,
    'second': NANOSECONDS_PER_SECOND,
    'seconds': NANOSECONDS_PER_SECOND,
    '″': NANOSECONDS_PER_SECOND,
    '"': NANOSECONDS_PER_SECOND,
    'ms': NANOSECONDS_PER_MILLISECOND,
    'millisecond': NANOSECONDS_PER_MILLISECOND,
    'milliseconds': NANOSECONDS_PER_MILLISECOND,
    'µs': NANOSECONDS_PER_MICROSECOND,
    'us': NANOSECONDS_PER_MICROSECOND,
    'microsecond': NANOSECONDS_PER_MICROSECOND,
    'microseconds': NANOSECONDS_PER_MICROSECOND,
    'ns': 1,
    'nanosecond': 1,
    'nanoseconds': 1,
}

__NANOSECONDS_TO_UNIT = {
    NANOSECONDS_PER_DAY: 'day',
    NANOSECONDS_PER_HOUR: 'hour',
    NANOSECONDS_PER_MINUTE: 'minute',
    NANOSECONDS_PER_SECOND: 'second',
    NANOSECONDS_PER_MILLISECOND: 'millisecond',
    NANOSECONDS_PER_MICROSECOND: 'microsecond',
    NANOSECONDS_PER_NANOSECOND: 'nanosecond',
}

__UU = [
    NANOSECONDS_PER_DAY,
    NANOSECONDS_PER_HOUR,
    NANOSECONDS_PER_MINUTE,
    NANOSECONDS_PER_SECOND,
]

__SPLIT_RE = re.compile(r'(?P<number>[.\d]+)(?P<unit>[^.\d]*)(?P<rest>.*)')

def parse_to_nanoseconds(s: str) -> int:
    """
    Parse a timestamp, returning nanoseconds.

    Some examples of accepted strings:

      123 days 5′ 10.3″
      1:23:45:57.39
      99
      99:59
      99:59.99
      99:59:59.999
      99:23:59:59.999
      123 Hours 5:10.3
      123 hours 10.25
      123 hours 99s1024
      1d 23 59 59
      1 23H 59 59
      1d 23 59 59s
      1 day 14 µs
    """
    s = s.strip()
    if not s:
        message = f'not a number: {s!r}'
        raise ValueError(message)

    # Split into parallel lists of number and unit.
    numbers: list[str] = []
    units: list[int] = []
    t = s
    while (m := __SPLIT_RE.fullmatch(t)):
        number, unit, t = m.group('number', 'unit', 'rest')
        numbers.append(number)
        units.append(parse_unit(unit))

    if t:
        message = f'not a number: {t!r}'
        raise ValueError(message)

    if all(units):
        pass
    elif any(units):
        units = __reconcile_units(s, numbers, units)
    else:
        units = __assign_units(s, numbers)

    total: float = 0
    for number, unit in zip(numbers, units, strict=True):
        total += float(number) * unit
    return int(total)

def __reconcile_units(s: str, numbers: list[str],
                      units: list[int]) -> list[int]:
    """Handle mixed cases."""
    units = list(units)
    length = len(units)

    # Find the rightmost unit.
    ri = -1
    for ri in range(length - 1, -1, -1):    # pragma: no branch
        if units[ri]:
            break
    if ri < 0:                              # pragma: no branch
        message = f'internal error: {s!r}'  # pragma: no coverage
        raise ValueError(message)           # pragma: no coverage

    if ri != length - 1:
        # There are units to the right to be filled in.
        u = units[ri]
        if ri == length - 2 and u == NANOSECONDS_PER_SECOND:
            # Special case: handle the use of ‘s’ as a decimal point.
            n = numbers[-1]
            if '.' in n:
                message = f'ambiguous decimal: {s!r}'
                raise ValueError(message)
            units[-1] = 10**(9 - len(n))
        else:
            fi = ri + 1
            try:
                ui = __UU.index(u) + 1
            except ValueError as e:
                message = f'ambiguous units: {s!r}'
                raise ValueError(message) from e
            ulength = len(__UU)
            while fi < length and ui < ulength:
                units[fi] = __UU[ui]
                fi += 1
                ui += 1
            if fi < length:
                message = f'too many numbers after {__NANOSECONDS_TO_UNIT[u]}s'
                raise ValueError(message)

    # Now work left.
    for i in range(ri, -1, -1):
        if units[i] == 0:
            u = units[ri]
            try:
                ui = __UU.index(u)
            except ValueError as e:
                message = f'ambiguous units: {s!r}'
                raise ValueError(message) from e
            if ui == 0:
                message = (f'no units larger than {__NANOSECONDS_TO_UNIT[u]}'
                           f'for {numbers[i]}')
                raise ValueError(message)
            units[i] = __UU[ui - 1]
        ri = i

    return units

def __assign_units(s: str, numbers: list[str]) -> list[int]:
    """Handle a colon-separated sequence."""
    if (length := len(numbers)) > 4:
        message = f'too many components: {s!r}'
        raise ValueError(message)
    if any('.' in n for n in numbers[:-1]):
        message = f'unexpected decimal: {s!r}'
        raise ValueError(message)
    for i, u, m in [(1, 'seconds', 60), (2, 'minutes', 60), (3, 'hours', 24)]:
        if length > i and float(numbers[-i]) >= m:
            message = f'{u} too large: {numbers[-i]!r}'
            raise ValueError(message)
    return __UU[-length :]

def parse_unit(s: str) -> int:
    s = s.strip().lower()
    if not s or s == ':':
        return 0
    if s in __UNIT_TO_NANOSECONDS:
        return __UNIT_TO_NANOSECONDS[s]
    message = f'not a unit: {s!r}'
    raise ValueError(message)



# SPDX-License-Identifier: MIT
"""Helper to format a repr resembling a constructor."""

from collections.abc import Iterable, Mapping
from typing import Any

# pylint: disable=dangerous-default-value
# pylint does not understand that Iterator and Mapping are immutable.
def mkrepr(
    obj: object,
    pos: Iterable[str],
    kws: Iterable[str] = [],
    defaults: Mapping[str, Any] = {},
) -> str:
    """
    Format a repr resembling a constructor.

    pos - positional argument names
    kws - keyword argument names
    defaults - argument default values
    """
    r = []
    kw = False
    for attr in pos:
        v = getattr(obj, attr)
        if v is None or v == defaults.get(attr):
            # Omitted, so subsequent parameters need a keyword.
            kw = True
        elif kw:
            r.append(f'{attr.lstrip("_")}={v!r}')
        else:
            r.append(repr(v))
    for attr in kws:
        v = getattr(obj, attr)
        if v is not None and v != defaults.get(attr):
            r.append(f'{attr.lstrip("_")}={v!r}')
    return f'{type(obj).__name__}({",".join(r)})'



# SPDX-License-Identifier: MIT
"""EAN13 - Vlju representing a EAN13."""

from fnattr.util import checksum
from fnattr.vlju.types.urn import URN

def is_valid_ean13(s: str) -> bool:
    """Check for 13-digit-only form."""
    return (len(s) == 13 and s.isdigit() and checksum.alt13(s[0 : 12]) == s[12])

def to13(s: str) -> str | None:
    """Convert other forms to E-13."""
    s = s.replace('-', '')
    s = s.replace('.', '')
    if len(s) == 8:
        s = '977' + s[: 7] + '000'  # ISSN → EAN13
    if len(s) == 9:
        s = '0' + s                 # SBN → ISBN
    if len(s) == 10:
        if s[0].lower() == 'm':     # noqa: SIM108
            s = '9790' + s[1 :]     # ISMN → EAN13
        else:
            s = '978' + s           # ISBN → EAN13
    if len(s) == 12:
        s = '0' + s                 # UPC-A → EAN13
    if len(s) == 13 and s[0 : 12].isdigit():
        return s[0 : 12] + checksum.alt13(s[0 : 12])
    return None

def key13(s: str) -> str:
    """Classify an E-13."""
    t = s[: 3]
    if t == '977':
        return 'issn'
    if t == '978':
        return 'isbn'
    if t == '979':
        if s[3] == '0':
            return 'ismn'
        return 'isbn'
    return 'ean13'

def as13(s: str, key: str) -> str | None:
    if (e := to13(s)) and (key13(e) == key):
        return e
    return None

class EAN13(URN):
    """Represents an EAN-13 article number."""

    def __init__(self, v: str, k: str = 'ean13') -> None:
        u = to13(v)
        if u is None:
            raise ValueError(v)
        super().__init__(u, k)

    def __str__(self) -> str:
        return self._value

    def __int__(self) -> int:
        return int(self._value)



# SPDX-License-Identifier: MIT
"""DOI - Document Object Identifier."""

import re
import urllib.parse

from collections.abc import Sequence
from typing import Self

from fnattr.util import escape
from fnattr.util.typecheck import needtype
from fnattr.vlju.types.info import Info
from fnattr.vlju.types.uri import URI, Authority
from fnattr.vlju.types.url import URL

class Prefix(list[int]):
    """Represents a DOI (or Handle) prefix."""

    def __init__(self, p: Self | Sequence[int] | str) -> None:
        if isinstance(p, str):
            p = list(map(int, p.split('.')))
        super().__init__(p)

    def is_doi(self) -> bool:
        return self[0] == 10

    def __str__(self) -> str:
        return '.'.join(map(str, self))

    @property
    def prefix(self) -> 'Prefix':
        # Implement prefix so that a Prefix can quack like a DOI.
        return self

class DOI(Info):
    """
    Represents a DOI (Document Object Identifier) or Handle.

    short:  sdoi
    long:   doi | uri
    where:
        sdoi    → `_prefix` ‘,’ `_suffix`
        doi     → ‘doi:’ `_prefix` ‘/’ `_suffix`ᵖ
    """

    _i = {'doi': Authority('doi'), 'hdl': Authority('hdl')}
    _u = {'doi': Authority('doi.org'), 'hdl': Authority('hdl.handle.net')}

    _matcher = re.compile(
        r"""
            (?P<scheme>
            |https?://((dx.)?doi.org|hdl.handle.net)/
            |(info:)?(hdl|doi)/
            |doi:/*
            )
            (?P<prefix>\d[\d.]*)
            [/,]
            (?P<suffix>.+)
            """,
        re.VERBOSE)

    def __init__(self, s: str | None = None, **kwargs) -> None:
        if s is None:
            prefix = Prefix(kwargs['prefix'])
            suffix = needtype(kwargs['suffix'], str)
        else:
            m = self._matcher.fullmatch(s)
            if not m:
                message = f'Not a DOI: {s}'
                raise ValueError(message)
            scheme, p, suffix = m.group('scheme', 'prefix', 'suffix')
            prefix = Prefix(p)
            if scheme.startswith(('http', 'info')):
                suffix = urllib.parse.unquote(suffix)
        # Note that DOI does not use Info path or authority.
        super().__init__('')
        self._prefix = prefix
        self._suffix = suffix.lower()
        self._kind = 'doi' if self._prefix.is_doi() else 'hdl'

    def prefix(self) -> Prefix:
        return self._prefix

    def suffix(self) -> str:
        return self._suffix

    def hdl(self, hdl: str = 'hdl') -> str:
        return f'info:{hdl}/{self.spath()}'

    def doi(self) -> str:
        if self._prefix.is_doi():
            return f'doi:{self._prefix}/{escape.path.encode(self._suffix)}'
        return super().lv()

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URI:
            return (self.spath(),
                    {
                        'scheme': 'info',
                        'sa': ':',
                        'authority': self._kind,
                        'query': self.query(),
                        'fragment': self.fragment(),
                    })
        if t is URL:
            return (self.spath(),
                    {
                        'scheme': 'https',
                        'authority': self._u[self._kind],
                        'query': self.query(),
                        'fragment': self.fragment(),
                    })
        raise self.cast_param_error(t)

    # URI overrides:

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DOI):
            return (self._prefix == other._prefix       # noqa: SLF001
                    and self._suffix == other._suffix)  # noqa: SLF001
        return False

    def authority(self) -> Authority:
        return self._i[self._kind]

    def sauthority(self) -> str:
        return self._kind

    def spath(self) -> str:
        return f'{self._prefix}/{self._suffix}'

    # Vlju overrides:

    def __str__(self) -> str:
        return f'{self._prefix},{self._suffix}'

    def lv(self) -> str:
        return self.doi()

    def __repr__(self) -> str:
        return f'DOI(prefix={self._prefix!r},suffix={self._suffix!r})'



# SPDX-License-Identifier: MIT
"""File - Local file path."""

import pathlib

from fnattr.util import escape
from fnattr.util.repr import mkrepr
from fnattr.vlju.types.uri import URI, Authority
from fnattr.vlju.types.url import URL

class File(URI):
    """Represents a local file path."""

    _authority = Authority('')

    def __init__(self, s: object = '') -> None:
        # `File` does not use the URI `path` component. Instead, it stores
        # the path as a `pathlib.Path` in `self._file`.
        if hasattr(s, 'cast_params'):
            s, _ = s.cast_params(type(self))
        if isinstance(s, pathlib.Path):
            self._file = s
            absolute = s.is_absolute()
        elif isinstance(s, str):
            self._file = pathlib.Path(s)
            absolute = s.startswith('/')
        else:
            raise TypeError(s)
        super().__init__(
            '',
            scheme='file',
            authority=self._authority,
            sa=('://' if absolute else ':'),
            ap='')

    # path() is already in use for the URI path component.
    def filename(self) -> pathlib.Path:
        return self._file

    def file(self) -> pathlib.Path:
        return self._file

    def cast_params(self, t: object) -> tuple[str, dict]:
        if t is URI or t is URL:
            return (str(self._file), {
                'scheme': 'file',
                'sa': self._sa,
                'ap': self._ap,
            })
        raise self.cast_param_error(t)

    # Vlju overrides:

    def __eq__(self, other: object) -> bool:
        if isinstance(other, File):
            return self._file == other._file    # noqa: SLF001
        return False

    def __str__(self) -> str:
        return str(self._file)

    def __repr__(self) -> str:
        return mkrepr(self, ['_file'])  # pragma: no cover

    # URI overrides:

    def path(self) -> str:
        return str(self._file)

    def spath(self) -> str:
        return escape.path.encode(str(self._file))



# SPDX-License-Identifier: MIT
"""Escape encodings."""

import shlex
import urllib.parse

from collections.abc import Callable, Mapping
from typing import NamedTuple

UNSAFE_ON_WINDOWS = '"*/:<>?\\|'
UNSAFE_ON_UNIX = '/'
UNSAFE_ON_MACOS = ':'

def mktrans_urlish(encode_chars: str) -> Mapping:
    return str.maketrans({
        c: ''.join(f'%{i:02X}' for i in c.encode('utf8'))
        for c in encode_chars + '%'
    })

def mkencode_urlish(encode_chars: str = '') -> Callable[[str], str]:
    tr = mktrans_urlish(encode_chars)

    def encode(s: str) -> str:
        return s.translate(tr)

    return encode

def mkencode_urllib(quote: Callable[[str, str], str],
                    safe: str = '') -> Callable[[str], str]:

    def encode(s: str) -> str:
        return quote(s, safe)

    return encode

class Escape(NamedTuple):
    """Escaping scheme with encode and decode operations."""

    encode: Callable[[str], str]
    decode: Callable[[str], str]

# Global registry of named Escapes.
escape: dict[str, Escape] = {}

def _register_escape(name: str, e: Escape) -> Escape:
    escape[name] = e
    return e

auth = _register_escape(
    'auth',
    Escape(
        mkencode_urllib(urllib.parse.quote, "!$&'()*+,;="),
        urllib.parse.unquote))

path = _register_escape(
    'path',
    Escape(
        mkencode_urllib(urllib.parse.quote, "!$&'()*+,;=:@/"),
        urllib.parse.unquote))

query = _register_escape(
    'query',
    Escape(
        mkencode_urllib(urllib.parse.quote_plus, ':@?/&='),
        urllib.parse.unquote_plus))

fragment = _register_escape('fragment', query)

unixfile = _register_escape(
    'unixfile', Escape(mkencode_urlish(UNSAFE_ON_UNIX), urllib.parse.unquote))

macfile = _register_escape(
    'macfile', Escape(mkencode_urlish(UNSAFE_ON_MACOS), urllib.parse.unquote))

umfile = _register_escape(
    'umfile',
    Escape(
        mkencode_urlish(UNSAFE_ON_UNIX + UNSAFE_ON_MACOS),
        urllib.parse.unquote))

winfile = _register_escape(
    'winfile', Escape(mkencode_urlish(UNSAFE_ON_WINDOWS), urllib.parse.unquote))

sh = _register_escape('sh',
                      Escape(shlex.quote, lambda s: ' '.join(shlex.split(s))))
